' Add a Reference to System.Configuration
Imports System.Configuration
Imports System.xml
Imports System.IO
Imports System.Text

'
' Allows for the editing of key/value pairs in the config file.  This required
' because the My.Settings values for the Application scope are read only, and 
' the values for User scope are recorded elsewhere.
' Note: You can only change a value, you can not add or delete a key
'
Public Class ConfigXML

    Private cfg As Configuration
    Private xd_app, xd_user As XmlDocument
    Private AssemblyName As String

    Public Enum Section As Integer
        Application = 0
        User
    End Enum

    '
    ' The default constructor loads the configuration file from the
    ' currently running application
    '
    Public Sub New()
        cfg = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None)
        AssemblyName = System.Reflection.Assembly.GetExecutingAssembly().GetName.Name

        LoadSections()
    End Sub

    '
    ' Load the configuration file for some other application
    '
    Public Sub New(ByVal ConfigPath As String, ByVal ExpectedAssemblyName As String)
        Dim map As New ExeConfigurationFileMap
        Dim s As ConfigurationSectionGroup

        If File.Exists(ConfigPath) Then
            map.ExeConfigFilename = ConfigPath
            cfg = ConfigurationManager.OpenMappedExeConfiguration(map, ConfigurationUserLevel.None)

            ' Get the Assembly name from the config file
            s = cfg.SectionGroups("applicationSettings")
            If IsNothing(s) Then
                Throw New ApplicationException("Configuration file missing, corrupt, or not the correct version")
            End If
            AssemblyName = s.Sections(0).SectionInformation.Name.Split("."c)(0)
            If AssemblyName <> ExpectedAssemblyName Then
                Throw New ApplicationException("Incorrect Configuration file" & vbCr & "Expecting configuration for " & ExpectedAssemblyName & ", however the configuration file is for " & AssemblyName)
            End If
        Else
            Throw New System.IO.FileNotFoundException
        End If

        LoadSections()
    End Sub

    '
    ' Save the changes back to the config file
    '
    Public Sub SaveSettings()
        Dim sec_app, sec_user As ConfigurationSection
        Dim sb As StringBuilder
        Dim xws As XmlWriterSettings
        Dim xw As XmlWriter

        ' Create settings for the XmlWriter to match the formatting of
        ' the *.config file
        xws = New XmlWriterSettings
        xws.Indent = True
        xws.IndentChars = "    "
        xws.OmitXmlDeclaration = True

        ' prep the XmlWriter
        sb = New StringBuilder
        xw = [XmlWriter].Create(sb, xws)
        If Not IsNothing(xd_app) Then
            xd_app.Save(xw)

            ' replace the entire applicationSettings section
            sec_app = CType(cfg.GetSection("applicationSettings/" & AssemblyName & ".My.MySettings"), ConfigurationSection)
            sec_app.SectionInformation.SetRawXml(sb.ToString)
        End If

        ' do it again for the userSettings section
        sb = New StringBuilder
        xw = [XmlWriter].Create(sb, xws)
        If Not IsNothing(xd_user) Then

            xd_user.Save(xw)
            sec_user = CType(cfg.GetSection("userSettings/" & AssemblyName & ".My.MySettings"), ConfigurationSection)
            sec_user.SectionInformation.SetRawXml(sb.ToString)
        End If

        ' save the changes to the config file
        cfg.Save()
    End Sub

    '
    ' Get a value from a key/value pair from the configuration file
    '
    Public Function GetValue(ByVal Sec As Section, ByVal Key As String) As String
        Dim xn As XmlNode

        xn = Nothing

        ' Use XPath navigation to find the proper node
        Select Case Sec
            Case Section.Application
                If IsNothing(xd_app) Then
                    Return Nothing
                End If
                xn = xd_app.SelectSingleNode("//setting[@name=""" & Key & """]")
            Case Section.User
                If IsNothing(xd_user) Then
                    Return Nothing
                End If
                xn = xd_user.SelectSingleNode("//setting[@name=""" & Key & """]")
        End Select

        If Not IsNothing(xn) Then
            ' is this a serialized string array?
            If xn.InnerXml.StartsWith("<value><ArrayOfString") Then
                Return XMLStringArraytoString(xn.InnerXml.Substring(7, xn.InnerXml.Length - 15))
            End If
            Return xn.InnerText
        End If

        Return Nothing
    End Function

    '
    ' Set a value in the key/value pair of the configuration file
    '
    Public Sub SetValue(ByVal Sec As Section, ByVal key As String, ByVal value As String)
        Dim xn As XmlNode

        xn = Nothing

        ' Use XPath navigation to find the proper node
        Select Case Sec
            Case Section.Application
                If IsNothing(xd_app) Then
                    Exit Sub
                End If
                xn = xd_app.SelectSingleNode("//setting[@name=""" & key & """]")
            Case Section.User
                If IsNothing(xd_user) Then
                    Exit Sub
                End If
                xn = xd_user.SelectSingleNode("//setting[@name=""" & key & """]")
        End Select

        If Not IsNothing(xn) Then
            If value.Contains(vbCr) Then
                ' We make the assumption that if there is an imbedded CR
                ' in the string, then the value is really a string array
                xn.InnerXml = "<value>" & vbCr & String2XMLStringArray(value) & vbCr & "        </value>"
            Else
                xn.InnerXml = "<value>" & value & "</value>"
            End If
        End If
    End Sub

    '
    ' Load the two XML documents with the appropriate section
    '
    Private Sub LoadSections()
        Dim sec_app, sec_user As ConfigurationSection

        ' get the applicationSettings section
        sec_app = CType(cfg.GetSection("applicationSettings/" & AssemblyName & ".My.MySettings"), ConfigurationSection)
        If Not IsNothing(sec_app) Then
            ' make an XmlDocument out of that section
            xd_app = New XmlDocument
            xd_app.LoadXml(sec_app.SectionInformation.GetRawXml)
        End If

        ' do it again for the userSettings section
        sec_user = CType(cfg.GetSection("userSettings/" & AssemblyName & ".My.MySettings"), ConfigurationSection)
        If Not IsNothing(sec_user) Then
            ' make an XmlDocument out of that section
            xd_user = New XmlDocument
            xd_user.LoadXml(sec_user.SectionInformation.GetRawXml)
        End If
    End Sub

    '
    ' Convert an XML serialized string array into a string
    '
    Private Shared Function XMLStringArraytoString(ByVal buf As String) As String
        Dim xmls As Serialization.XmlSerializer
        Dim sr As StringReader
        Dim temp() As String

        ' build a serializer that handles string arrays
        xmls = New System.Xml.Serialization.XmlSerializer(GetType(String()))

        ' do the deed
        sr = New StringReader(buf)
        temp = CType(xmls.Deserialize(sr), String())
        If temp.Length = 0 Then
            Return ""
        End If

        Return Join(temp, vbCrLf)
    End Function

    '
    ' Convert a string array into an XML serialized string
    '
    Private Shared Function String2XMLStringArray(ByVal buf As String) As String
        Dim xmls As Serialization.XmlSerializer
        Dim sb As StringBuilder
        Dim xw As XmlWriter
        Dim xws As XmlWriterSettings
        Dim temp() As String

        ' trim any extra blank lines at the end
        buf = buf.TrimEnd(vbCrLf.ToCharArray)

        ' build a string array
        temp = buf.Split(vbCr.ToCharArray)
        For i As Integer = 0 To UBound(temp)
            temp(i) = temp(i).Trim(vbLf.ToCharArray)
        Next

        ' setup the formatting options
        xws = New XmlWriterSettings
        xws.Indent = True
        xws.IndentChars = "    "
        xws.OmitXmlDeclaration = True

        ' prep the XmlWriter
        sb = New System.Text.StringBuilder
        xw = System.Xml.XmlWriter.Create(sb, xws)

        ' build a serializer that handles string arrays
        xmls = New System.Xml.Serialization.XmlSerializer(GetType(String()))

        ' do the deed
        xmls.Serialize(xw, temp)

        ' add a bit of padding to each line
        temp = sb.ToString.Split(vbCr.ToCharArray)
        For i As Integer = 0 To UBound(temp)
            temp(i) = "            " & temp(i).Trim(vbLf.ToCharArray)
        Next

        Return Join(temp, vbCr)
    End Function
End Class
