' Add a Reference to System.Management
' Add a Reference to System.DirectoryServices.dll
' Add a COM Reference to Windows Script Host Object Model
Imports System.Management
Imports System.DirectoryServices
Imports System.IO
Imports System.Globalization
Imports System.Runtime.InteropServices
Imports System.Collections.Specialized

' Use these compile-time variables to reduce the amount of data collected
#Const TRIM_PHANTOM_MONITORS = True
#Const TRIM_PHANTOM_MICE = True
#Const TRIM_PHANTOM_VIDEOS = True
#Const TRIM_DISABLED_SYSDRIVERS = True
#Const TRIM_TAMPER_ATTEMPTS = True
#Const TRIM_SUCCESS_PRIVILEGE_USE = True

' Should the system drive letter be used in the formulation of the ID field?
' Useful if users who have multiple boot partitions with the same PC name.
#Const USE_DRIVE_LETTER_IN_ID = False
'
' The main body of the application.  The snoop class gathers the data in a
' detached dataset.  Most errors are captured and included as records in the
' database and recorded to an error log (See Misc.ErrorLog for details)
'
Public Class Snoop
    Implements IDisposable

    Private RemotePC As String              ' Blank for local PC, or the remote PC's name
    Private ds As SOSOSDataSet              ' the detached dataset to store our info
    Private MachineName As String           ' the true name of the PC (not an IP address)
    Private ID As String                    ' the primary key (PC name and MAC address)
    Private IsWin32 As Boolean              ' is the PC at least WinNT?
    Private OSVer As Double                 ' the operating system version (5.1= WinXP)
    Private DomainRole As Integer           ' Workgroup, Domain member, Domain Controller
    Private SystemDrive As String           ' C:
    Private SystemRoot As String            ' C:\Windows
    Private SystemDirectory As String       ' C:\Windows\System32
    Private ProgramFiles As String          ' C:\Program Files
    Private ProgramFilesx86 As String       ' C:\Program Files (x86)
    Private CommonProgramFiles As String    ' C:\Program Files\Common Files
    Private CommonProgramFilesx86 As String ' C:\Program Files\Common Files (x86)
    Private WinDir As String                ' C:\Windows
    Private AllUsersProfile As String       ' C:\Documents and Settings\All Users
    Private Is64bit As Boolean              ' A 64-bit OS?
    Private wmi As ManagementClass
    Private reg_hklm As RegistryKey         ' HKEY_LOCAL_MACHINE
    Private reg_hkcr As RegistryKey         ' HKEY_CURRENT_USER
    Private reg_hku As RegistryKey          ' HKEY_USERS

    Private UID As String                   ' The User ID (the name) of the current user
    Private SID As String                   ' The Security ID for the PC
    Private Netcard(-1) As Integer
    Private culture As CultureInfo

    ' used in GetProfiles()
    <StructLayout(LayoutKind.Explicit)> _
    Private Structure LongDate
        <FieldOffset(0)> Dim Long64 As Int64
        <FieldOffset(0)> Dim date_lo As Int32
        <FieldOffset(4)> Dim date_hi As Int32
    End Structure

    Private disposed As Boolean
    Private ProfileList As StringDictionary
    Private SystemPathVars As StringDictionary

    Public Sub New(ByVal PC_Name As String, ByVal SOSOSds As SOSOSDataset)
        ' a quick sanity check...
        If IsNothing(SOSOSds) Then
            Throw New ApplicationException("DataSet parameter can not be null")
        End If

        ' Copy the passed parameters
        RemotePC = PC_Name
        ds = SOSOSds

        ' Initialize a bunch of variables that we will need throughout
        ' the remainder of the program
        MachineName = ""
        ID = ""
        IsWin32 = False
        OSVer = 0.0
        DomainRole = 0
        SystemDrive = "C:"
        SystemRoot = "C:\Windows"
        SystemDirectory = "C:\Windows\System32"
        ProgramFiles = "C:\Program Files"
        ProgramFilesx86 = "C:\Program Files (x86)"
        CommonProgramFiles = "C:\Program Files\Common Files"
        CommonProgramFilesx86 = "C:\Program Files (x86)\Common Files"
        WinDir = "C:\Windows"
        AllUsersProfile = "C:\Documents and Settings\All Users"
        Is64bit = False
        wmi = Nothing
        reg_hkcr = Nothing
        reg_hklm = Nothing
        reg_hku = Nothing
        SID = ""
        UID = ""

        ' Set the culture information for formating dates
        culture = New CultureInfo("en-US", True)

        ' Enable some privileges that are required to read the security event
        ' log and mount/unmount registry hives.  This will fail if the current
        ' user is not an administrator, but we handle that elsewhere.
        Misc.SetPrivileges()

        ' Create a dictionary to hold the list of registry "mount points"
        ' and their associated ProfileNames
        ProfileList = New StringDictionary
        ' Create another dictionary for expanding Environmental Variables from
        ' the RemotePC's perspective
        SystemPathVars = New StringDictionary

        ' prime the pump...
        GetInitialData()
    End Sub

    ' SOSOS requires a bunch of "base data" that is used throughout the rest of the
    ' program.  This class will create the base data and open a few handles for the
    ' registry and WMI.

    ' It also does some of the initial "sanity checking" to make sure we're not
    ' gonna have problems later on in the program.  Therefore, exceptions are NOT
    ' captured or processed at this level.

    Private Sub GetInitialData()
        Dim obj As ManagementObject
        Dim UserName, MgmtPath, temp, parts(), MAC, Manufacturer As String
        Dim LocalAdmin As Boolean
        Dim key As RegistryKey
        Dim cnt As Integer
        Dim dr As SOSOSDataset.SOSRow

        ' This is just a deliberate connection to a remote PC to see if we
        ' get any errors.  It takes about 7 seconds on WinXP to "time out"
        If RemotePC <> "" Then
            Dim scope As ManagementScope
            scope = New ManagementScope("\\" & RemotePC & "\root\cimv2")
            scope.Options.Authentication = AuthenticationLevel.Packet
            scope.Options.Impersonation = ImpersonationLevel.Impersonate
            scope.Connect()
        End If

        ' open the registry keys (local or remote)
        If RemotePC <> "" Then
            reg_hklm = RegistryKey.OpenRemoteBaseKey(RegistryHive.LocalMachine, RemotePC)
            reg_hkcr = RegistryKey.OpenRemoteBaseKey(RegistryHive.ClassesRoot, RemotePC)
            reg_hku = RegistryKey.OpenRemoteBaseKey(RegistryHive.Users, RemotePC)
        Else
            reg_hklm = Registry.LocalMachine
            reg_hkcr = Registry.ClassesRoot
            reg_hku = Registry.Users
        End If

        UserName = ""
        MgmtPath = ""
        If RemotePC <> "" Then
            MgmtPath = "\\" & RemotePC & "\root\cimv2:"
        End If

        ' save this as the "root" of the other WMI calls
        wmi = New ManagementClass(MgmtPath & "Win32_ComputerSystem")

        ' Connect to the WMI repository, get basic computer info.
        For Each obj In wmi.GetInstances()
            ' The MachineName could be different from RemotePC (ie,
            ' RemotePC could be an IP address)
            MachineName = obj("Name").ToString
            ' Current User (or null if nobody is logged in)
            If Not IsNothing(obj("UserName")) Then
                UserName = obj("UserName").ToString
            End If
            ' Is this PC a member of a domain?
            If Not IsNothing(obj("DomainRole")) Then
                DomainRole = Convert.ToInt32(obj("DomainRole"))
            End If

            ' Is this a 64-bit OS?
            If obj("SystemType").ToString.IndexOf("64") >= 0 Then
                Is64bit = True
            End If
        Next

        ' Get the basic WMI information about the Operating System
        wmi.Path.RelativePath = "Win32_OperatingSystem"
        For Each obj In wmi.GetInstances()
            SystemRoot = obj("WindowsDirectory").ToString
            SystemDrive = Left(SystemRoot, 2)
            If Not IsNothing(obj("SystemDirectory")) Then
                SystemDirectory = obj("SystemDirectory").ToString
            End If
            ' just one decimal point please
            parts = Split(obj("Version").ToString, ".")
            If parts.GetUpperBound(0) >= 1 Then
                OSVer = Double.Parse(parts(0) & "." & parts(1), CultureInfo.InvariantCulture)
            End If

            ' is this OS at least WinNT
            If Not IsNothing(obj("OSType")) AndAlso Convert.ToInt32(obj("OSType")) = 18 Then
                IsWin32 = True
            End If
        Next

        ' Get the MAC address so we can build a unique identifier.
        ' If there is more than one Network card, use only the first
        ' one's MAC (to avoid confusion of multiple ID's)
        MAC = ""
        Manufacturer = ""
        wmi.Path.RelativePath = "Win32_NetworkAdapter"
        For Each obj In wmi.GetInstances()
            Manufacturer = CType(obj("Manufacturer"), String)
            If Not IsNothing(obj("MACAddress")) AndAlso Not obj("Name").ToString.StartsWith("WAN") And Manufacturer <> "Microsoft" Then
                MAC = obj("MACAddress").ToString
                ' If multiple cards, use only the first one.  This might
                ' not be the card "in use", but at least we're consistent
                Exit For
            End If
        Next

        ' In rare cases, the above doesn't work.  So we try this instead
        If MAC = "" Then
            wmi.Path.RelativePath = "Win32_NetworkAdapterConfiguration"
            For Each obj In wmi.GetInstances()
                If Not IsNothing(obj("MACAddress")) Then
                    MAC = obj("MACAddress").ToString
                    Exit For
                End If
            Next
        End If

        ' The ID must be unique (for the sake of the database), so it's
        ' the combination of the machine name, the boot driver letter, and
        ' the MAC address.
        ' Note: That makes a horribly long primary key for the database,
        ' but too bad!
#If USE_DRIVE_LETTER_IN_ID Then
        ID = MachineName & SystemDrive & MAC
#Else
        If MAC = "" Then
            ID = MachineName
        Else
            ID = MachineName & "(" & MAC & ")"
        End If
#End If

        ' Do a simple "probing" of the filesystem as as test.  We don't want
        ' any surprises later on.  Failure is expected on remote Win9x systems,
        ' so don't panic
        Try
            Dim TempDate As Date
            TempDate = File.GetLastWriteTime(Misc.ConvUNC(RemotePC, SystemRoot))
        Catch ex As Exception
            If IsWin32 Then
                Throw New ApplicationException("Can't get read from Administrative share", ex)
            End If
        End Try

        ' Get the user ID of the person currently logged on to the system.
        LocalAdmin = False
        If RemotePC <> "" Then
            ' This is one of those strange quirks... if the user on the 
            ' remote PC is not an administrator, then the UserName comes
            ' up blank, so we try this other method.
            If UserName = "" Then
                UserName = Misc.GetRemoteUser(MachineName, reg_hku)
            End If
        Else
            ' We started with what WMI provided above... but Remote Desktop
            ' (aka Terminal Services) gives the wrong answer, it gives the
            ' name of user logged into the console!
            If Environment.OSVersion.Platform = PlatformID.Win32NT AndAlso Misc.IsRemoteSession() Then
                UserName = Environment.UserDomainName & "\" & Environment.UserName
            End If
        End If

        If UserName <> "" Then
            ' clean it up a bit
            cnt = InStr(UserName, "\")
            If cnt > 0 Then
                UID = UserName
                UserName = Mid(UserName, cnt + 1)
            Else
                ' for Win9x, build the UID
                UID = MachineName & "\" & UserName
            End If

            ' Are you an Administrator on this PC?
            If IsWin32 Then
                If RemotePC <> "" Then
                    ' See GetAdmins for the (poor) WinNT technique
                    If OSVer > 4.0 Then
                        LocalAdmin = Misc.IsRemoteAdmin(MachineName, reg_hku, UID)
                    End If
                Else
                    LocalAdmin = Misc.IsAdmin()
                End If
            End If

            ' The UID value is used later in the GetUserLogin method
            UID = UID.ToUpper
        End If

        ' Build the System Path Variables dictionary.  We use these to expand
        ' environmental variables that come from registry values on a remote
        ' PC (normally the GetValue() function would expand those variables 
        ' using *our* environment).
        SystemPathVars.Add("systemdrive", SystemDrive)
        SystemPathVars.Add("systemroot", SystemRoot)

        ' Get the remainder of the System Path Variables
        key = reg_hklm.OpenSubKey("SOFTWARE\Microsoft\Windows\CurrentVersion", False)
        If Not IsNothing(key) Then
            ' If this key is being redirected by a x64 OS, it already has the
            ' " (x86)" string tacked onto the end, so we need to "undo" that
            ' fix and get back to the real directories.
            ProgramFiles = key.GetValue("ProgramFilesDir", ProgramFiles).ToString.Replace(" (x86)", "")
            CommonProgramFiles = key.GetValue("CommonFilesDir", CommonProgramFiles).ToString.Replace(" (x86)", "")
            ProgramFilesx86 = ProgramFiles & " (x86)"
            CommonProgramFilesx86 = Path.GetDirectoryName(CommonProgramFiles) & " (x86)\" & Path.GetFileName(CommonProgramFiles)
            key.Close()
        End If

        If Not IsWin32 Then
            AllUsersProfile = SystemRoot & "\All Users"
        End If
        key = reg_hklm.OpenSubKey("SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProfileList", False)
        If Not IsNothing(key) Then
            Dim ProfilesDirectory As String

            ' Vista uses ProgramData in lieu of AllUsersProfile
            temp = key.GetValue("ProgramData", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString
            AllUsersProfile = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
            If AllUsersProfile = "" Then
                AllUsersProfile = key.GetValue("AllUsersProfile", "All Users").ToString
            End If

            ' Get the Profiles directory
            If OSVer < 5.0 Then
                ' WinNT doesn't have this key, so we fake it
                ProfilesDirectory = SystemRoot & "\Profiles"
            Else
                temp = key.GetValue("ProfilesDirectory", "C:\Documents and Settings", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString
                ProfilesDirectory = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
            End If
            If AllUsersProfile.Length > 2 AndAlso AllUsersProfile.Substring(1, 1) <> ":" Then
                AllUsersProfile = ProfilesDirectory & "\" & AllUsersProfile
            End If

            key.Close()
        End If

        ' Add the remainder of the System Path Variables
        SystemPathVars.Add("programfiles", ProgramFiles)
        SystemPathVars.Add("programfilesx86", ProgramFilesx86)
        SystemPathVars.Add("programfiles(x86)", ProgramFilesx86)
        SystemPathVars.Add("commonprogramfiles", CommonProgramFiles)
        SystemPathVars.Add("commonprogramfilesx86", CommonProgramFilesx86)
        SystemPathVars.Add("commonprogramfiles(x86)", CommonProgramFilesx86)
        SystemPathVars.Add("windir", SystemRoot)
        SystemPathVars.Add("allusersprofile", AllUsersProfile)

        ' Create the prototype record for the SOSOS table.  Since the foreign
        ' key constraints are in place, we must populate the tables in the
        ' correct order
        dr = ds.SOS.NewSOSRow
        dr.ID = ID
        dr.PC_Name = MachineName
        dr.UserName = UserName.ToLower
        dr.IsAdmin = LocalAdmin
        dr.CurrentDate = Now()
        dr.DHCP = False
        dr.Net_Services = False
        dr.CDROM = False
        dr.Modem = False
        dr.SoundCard = False
        dr.PCMCIA = False
        dr.SCSI = False
        dr.SmartCard = False
        dr.USB = False
        dr.AutoUpdate = False
        ds.SOS.AddSOSRow(dr)

    End Sub

    '
    ' Get the Local Account data
    '
    Public Sub GetAccounts()
        ' Collect data only if this not a Win9x client and is NOT a domain
        ' controller (DCs don't have local accounts!)
        If DomainRole < 4 And IsWin32 Then
            Dim dr As SOSOSDataset.AccountsRow

            Try
                ' We use two different techniques here, depending on if the PC
                ' is in a domain or not.  We use WMI for workgroups, and ADSI
                ' for domains (and WinVista).
                If (DomainRole = 0 Or DomainRole = 2) And OSVer < 6.0 Then
                    Dim obj As ManagementObject

                    wmi.Path.RelativePath = "Win32_UserAccount"
                    For Each obj In wmi.GetInstances()
                        dr = ds.Accounts.NewAccountsRow
                        dr.ID_Account = ID
                        dr.Account = Left(obj("Name").ToString, ds.Accounts.AccountColumn.MaxLength)
                        If Not IsNothing(obj("FullName")) AndAlso obj("FullName").ToString.Length > 0 Then
                            dr.Account_FullName = Left(obj("FullName").ToString, ds.Accounts.Account_FullNameColumn.MaxLength)
                        End If
                        If Not IsNothing(obj("Description")) AndAlso obj("Description").ToString.Length > 0 Then
                            dr.Account_Description = Left(obj("Description").ToString, ds.Accounts.Account_DescriptionColumn.MaxLength)
                        End If
                        dr.Account_Disabled = CType(obj("Disabled"), Boolean)
                        dr.Account_LockedOut = CType(obj("Lockout"), Boolean)
                        dr.Account_PwdRequired = CType(obj("PasswordRequired"), Boolean)
                        dr.Account_PwdChangeable = CType(obj("PasswordChangeable"), Boolean)
                        dr.Account_PwdExpires = CType(obj("PasswordExpires"), Boolean)

                        ' Grab the administrator's SID for later use in the
                        ' PC_SID table. 
                        If Not IsNothing(obj("SID")) AndAlso obj("SID").ToString.EndsWith("-500") Then
                            SID = obj("SID").ToString.Substring(0, obj("SID").ToString.Length - 4)
                        End If
                        ds.Accounts.AddAccountsRow(dr)
                    Next
                Else    ' use ADSI for domain members
                    Dim acct, accts As DirectoryEntry
                    Dim temp As String
                    Dim i As Integer

                    accts = New DirectoryEntry("WinNT://" & MachineName & ",Computer")
                    For Each acct In accts.Children
                        If acct.SchemaClassName = "User" Then
                            dr = ds.Accounts.NewAccountsRow
                            dr.ID_Account = ID
                            dr.Account = Left(acct.Name, ds.Accounts.AccountColumn.MaxLength)
                            If Not IsNothing(acct.Properties("FullName")) AndAlso acct.Properties("FullName").Value.ToString.Length > 0 Then
                                dr.Account_FullName = Left(acct.Properties("FullName").Value.ToString, ds.Accounts.Account_FullNameColumn.MaxLength)
                            End If
                            If Not IsNothing(acct.Properties("Description")) AndAlso acct.Properties("Description").Value.ToString.Length > 0 Then
                                dr.Account_Description = Left(acct.Properties("Description").Value.ToString, ds.Accounts.Account_DescriptionColumn.MaxLength)
                            End If
                            i = CInt(acct.Properties("UserFlags").Value)
                            dr.Account_Disabled = CBool(i And 2)
                            dr.Account_LockedOut = CBool(i And 16)
                            dr.Account_PwdRequired = Not CBool(i And 32)
                            ' Note: ADS_UF_PASSWD_CANT_CHANGE is only supported
                            ' in a domain environment
                            dr.Account_PwdChangeable = Not CBool(i And 64)
                            dr.Account_PwdExpires = Not CBool(i And 65536)

                            ' Grab the administrator's SID for later use in the 
                            ' PC_SID table.
                            If Not IsNothing(acct.Properties("objectSID")) Then
                                temp = Misc.SIDtoString(CType(acct.Properties("objectSID").Value, Byte()))
                                If temp.EndsWith("-500") Then
                                    SID = temp.Substring(0, temp.Length - 4)
                                End If
                            End If
                            ds.Accounts.AddAccountsRow(dr)
                        End If
                    Next
                End If
            Catch ex As Exception
                dr = ds.Accounts.NewAccountsRow
                dr.ID_Account = ID
                dr.Account_Disabled = False
                dr.Account_LockedOut = False
                dr.Account_PwdRequired = False
                dr.Account_PwdChangeable = False
                dr.Account_PwdExpires = False
                dr.Account = Left("Error: " & ex.Message, ds.Accounts.AccountColumn.MaxLength)
                ds.Accounts.AddAccountsRow(dr)
                Misc.ErrorLog(ex, MachineName, "Account", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the list of accounts in the local Administrators Group.
    '
    Public Sub GetAdmins()
        ' Win9x clients don't have Groups!
        If IsWin32 Then
            Dim oq As ObjectQuery
            Dim mos As ManagementObjectSearcher
            Dim obj As ManagementObject
            Dim ts As TimeSpan
            Dim eo As EnumerationOptions
            Dim DomainName, temp, Name, Domain As String
            Dim dr As SOSOSDataset.AdminsRow

            Try
                ' This is a bit strange... we can't be sure if the Domain field
                ' of the Win32_Group instance for "Administrators" is gonna be
                ' "BUILTIN" or the name of PC.  So, we do a query to see which
                ' one it is
                oq = New ObjectQuery("SELECT * FROM Win32_Group WHERE (Domain='BUILTIN' OR Domain='" & MachineName & "') AND Name='Administrators'")
                mos = New ManagementObjectSearcher(wmi.Scope, oq)
                DomainName = ""
                For Each obj In mos.Get()
                    DomainName = obj("Domain").ToString
                    ' This query will produce duplicates!
                    Exit For
                Next

                ' OK, now that we know the value for 'Domain', we can enumerate
                ' the user in the group using the Win32_GroupUser association.
                oq = New ObjectQuery("Select PartComponent from Win32_GroupUser where GroupComponent=""Win32_Group.Domain='" & DomainName & "',Name='Administrators'""")
                ' This sometimes takes a very long time when run on a domain
                ' controller, so we stop it after 30 seconds and just hope that
                ' we're satisfied with whatever data we've collected so far.
                ts = New TimeSpan(0, 0, 0, 30, 0)
                eo = New EnumerationOptions(Nothing, ts, 1, False, True, False, False, False, False, False)

                Try
                    mos = New ManagementObjectSearcher(wmi.Scope, oq, eo)
                    For Each obj In mos.Get()
                        dr = ds.Admins.NewAdminsRow
                        dr.ID_Admin = ID
                        temp = obj("PartComponent").ToString

                        ' Instead of using Associations to create the object,
                        ' we just parse the WMI string to extract the Domain
                        ' and Name (must faster)
                        Domain = Split(temp, Chr(34))(1)
                        Name = Split(temp, Chr(34))(3)
                        dr.Admin_UserID = Domain & "\" & Name

                        ' parse it again to figure out the type
                        If temp.IndexOf("Win32_Group") >= 0 Then
                            dr.Admin_AcctType = "Group"
                        Else
                            dr.Admin_AcctType = "User"
                        End If
                        ds.Admins.AddAdminsRow(dr)

                        ' This is a crude way to deterine if a remote WinNT
                        ' account is in the Administrators group.  It's not
                        ' perfect since it doesn't test for nested group,
                        ' membership, but it's better than nothing!
                        If RemotePC <> "" And OSVer = 4.0 And IsWin32 Then
                            If UID = dr.Admin_UserID.ToUpper And dr.Admin_AcctType = "User" Then
                                ds.SOS(0).IsAdmin = True
                            End If
                        End If
                    Next
                Catch sme As System.Management.ManagementException
                    ' just ignore a "Timed Out" error... but let all others
                    ' errors bubble up
                    If sme.ErrorCode <> ManagementStatus.Timedout Then
                        Throw
                    End If
                End Try

                ' Sort the list to make it pretty
                Misc.DataTableSort(ds.Tables("Admins"), "Admin_UserID")
            Catch ex As Exception
                dr = ds.Admins.NewAdminsRow
                dr.ID_Admin = ID
                dr.Admin_UserID = Left("Error: " & ex.Message, ds.Admins.Admin_UserIDColumn.MaxLength)
                ds.Admins.AddAdminsRow(dr)
                Misc.ErrorLog(ex, MachineName, "Admins", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the Windows Update settings
    '
    Public Sub GetAutoUpdate()
        Dim n As Integer
        Dim key As RegistryKey
        Dim temp As String
        Dim dr As SOSOSDataset.AutoUpdateRow

        Try
            dr = ds.AutoUpdate.NewAutoUpdateRow
            dr.ID_AutoUpdate = ID
            dr.AutoUpdate_Enabled = False
            dr.AutoUpdate_UseSUS = False

            ' this key has the Windows Update settings
            key = reg_hklm.OpenSubKey("SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate\AU", False)
            If Not IsNothing(key) Then
                n = CInt(key.GetValue("NoAutoUpdate", 0))
                dr.AutoUpdate_Enabled = Not CBool(n)
                n = CInt(key.GetValue("UseWUServer", 0))
                dr.AutoUpdate_UseSUS = CBool(n)
                n = CInt(key.GetValue("AUOptions", 0))
                dr.AutoUpdate_Options = Choose(n + 1, "Unknown", "Disabled", "Notify Only", "Download", "Auto Install").ToString
                key.Close()
            End If

            key = reg_hklm.OpenSubKey("SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate", False)
            If Not IsNothing(key) Then
                dr.AutoUpdate_Server = Left(CStr(key.GetValue("WUServer", "")), ds.AutoUpdate.AutoUpdate_ServerColumn.MaxLength)
                key.Close()
            End If

            ' this key has the current status (for the older SUS client)
            key = reg_hklm.OpenSubKey("SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update", False, RegWow64Options.KEY_WOW64_64KEY)
            If Not IsNothing(key) Then
                ' Use this key if you're not using SUS
                If dr.AutoUpdate_Enabled = False Then
                    n = CInt(key.GetValue("AUOptions", 0))
                    If n > 1 Then
                        dr.AutoUpdate_Enabled = True
                    End If
                    dr.AutoUpdate_Options = CStr(Choose(n + 1, "Unknown", "Disabled", "Notify Only", "Download", "Auto Install"))
                End If
                n = CInt(key.GetValue("AUState", 9))
                dr.AutoUpdate_State = CStr(Choose(n + 1, "Initializing", "Wizard", "Detect Pending", "Download Pending", "Download in Progress", "Install Pending", "Install in Progress", "Disabled", "Reboot Pending", "Error"))
                temp = key.GetValue("LastWaitTimeout", "").ToString
                If IsDate(temp) Then
                    dr.AutoUpdate_NextRun = [Date].Parse(temp, culture)
                Else
                    temp = key.GetValue("NextDetectionTime", "").ToString
                    If IsDate(temp) Then
                        dr.AutoUpdate_NextRun = [Date].Parse(temp, culture)
                    End If
                End If
                key.Close()
            End If

            ' The new WSUS client doesn't support the AUState registry
            ' value, so we have to read the log file instead
            If dr.IsAutoUpdate_StateNull OrElse dr.AutoUpdate_State = "Error" Then
                Dim s As Stream
                Dim sr As StreamReader
                Dim LogFile, parts(), buf As String

                LogFile = Misc.ConvUNC(RemotePC, SystemRoot & "\SoftwareDistribution\ReportingEvents.log")

                If File.Exists(LogFile) Then
                    ' read the file, retaining the last relevant line
                    s = File.Open(LogFile, FileMode.Open, FileAccess.Read, FileShare.ReadWrite)
                    sr = New StreamReader(s, True)
                    Do While Not sr.Peek < 0
                        buf = sr.ReadLine()
                        parts = buf.Split(vbTab.ToCharArray)
                        ' They changes the format of the log file during one of the updates/patches
                        ' so we check to see if the array size is 12 or 13 
                        Select Case parts.Length
                            Case 12
                                If parts(3) <> "148" And parts(5) <> "{00000000-0000-0000-0000-000000000000}" Then
                                    dr.AutoUpdate_State = Left(parts(8) & ", " & parts(9) & ", " & parts(10), ds.AutoUpdate.AutoUpdate_StateColumn.MaxLength)
                                End If
                            Case 13
                                If parts(4) <> "148" And parts(6) <> "{00000000-0000-0000-0000-000000000000}" Then
                                    dr.AutoUpdate_State = Left(parts(9) & ", " & parts(10) & ", " & parts(11), ds.AutoUpdate.AutoUpdate_StateColumn.MaxLength)
                                End If
                        End Select
                    Loop
                    sr.Close()
                End If
            End If

            ' We mark the SOS summary item as "true" only if all three of the
            ' following conditions are true:
            '   a) the client is using the SUS server in the config file
            '   b) SUS is enabled
            '   c) the client is set to use a SUS server
            If Not dr.IsAutoUpdate_ServerNull AndAlso ([String].Compare(dr.AutoUpdate_Server, My.Settings.SUS_Server, True) = 0 Or dr.AutoUpdate_Server.ToUpper Like My.Settings.SUS_Server.ToUpper) Then
                If dr.AutoUpdate_Enabled = True And dr.AutoUpdate_UseSUS = True Then
                    ds.SOS(0).AutoUpdate = True
                End If
            End If

            ds.AutoUpdate.AddAutoUpdateRow(dr)
        Catch ex As Exception
            dr = ds.AutoUpdate.NewAutoUpdateRow
            dr.ID_AutoUpdate = ID
            dr.AutoUpdate_Enabled = False
            dr.AutoUpdate_UseSUS = False
            dr.AutoUpdate_LastQFE = Left("Error: " & ex.Message, ds.AutoUpdate.AutoUpdate_LastQFEColumn.MaxLength)
            ds.AutoUpdate.AddAutoUpdateRow(dr)
            Misc.ErrorLog(ex, MachineName, "AutoUpdate", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the Computer's BIOS information
    '
    Public Sub GetBIOS()
        Dim obj As ManagementObject
        Dim dr As SOSOSDataset.BIOSRow

        Try
            wmi.Path.RelativePath = "Win32_BIOS"
            For Each obj In wmi.GetInstances()
                dr = ds.BIOS.NewBIOSRow
                dr.ID_BIOS = ID
                If Not IsNothing(obj("Manufacturer")) Then
                    dr.BIOS = Left(obj("Manufacturer").ToString, ds.BIOS.BIOSColumn.MaxLength)
                End If
                ' we prefer the SMB BIOS version, if available
                If CBool(obj("SMBIOSPresent")) Then
                    dr.BIOS_Version = Left(obj("SMBIOSBIOSVersion").ToString, ds.BIOS.BIOS_VersionColumn.MaxLength)
                Else
                    dr.BIOS_Version = Left(obj("Version").ToString, ds.BIOS.BIOS_VersionColumn.MaxLength)
                End If
                dr.BIOS_Date = ManagementDateTimeConverter.ToDateTime(obj("releasedate").ToString)
                ds.BIOS.AddBIOSRow(dr)
            Next
        Catch ex As Exception
            dr = ds.BIOS.NewBIOSRow
            dr.ID_BIOS = ID
            dr.BIOS = Left("Error: " & ex.Message, ds.BIOS.BIOSColumn.MaxLength)
            ds.BIOS.AddBIOSRow(dr)
            Misc.ErrorLog(ex, MachineName, "BIOS", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the "Components"... Similar to software but considered optional parts
    ' of the operating system
    '
    Public Sub GetComponents()
        Dim cnt As Integer
        Dim temp, Component, ComponentName As String
        Dim key, subkey As RegistryKey
        Dim dr As SOSOSDataset.ComponentsRow

        Try
            ' pick a large number so resorting won't create duplicate numbers
            cnt = 10000
            key = reg_hklm.OpenSubKey("SOFTWARE\Microsoft\Active Setup\Installed Components", False)
            If Not IsNothing(key) Then
                For Each temp In key.GetSubKeyNames()
                    subkey = key.OpenSubKey(temp, False)
                    If subkey.ValueCount > 0 Then
                        Component = ""
                        ComponentName = ""
                        ' look at both the default value and the ComponentID value
                        If Not IsNothing(subkey.GetValue("")) Then
                            Component = Left(subkey.GetValue("").ToString, ds.Components.ComponentColumn.MaxLength)
                        End If
                        If Not IsNothing(subkey.GetValue("ComponentID")) Then
                            ComponentName = Left(subkey.GetValue("ComponentID").ToString, ds.Components.Component_NameColumn.MaxLength)
                        End If
                        ' record this entry only if we can get a valid name
                        If Component <> "" Or ComponentName <> "" Then
                            dr = ds.Components.NewComponentsRow
                            dr.ID_Component = ID
                            dr.Component_ID = cnt
                            dr.Component = Component
                            dr.Component_Name = ComponentName
                            dr.Component_Version = ""
                            If Not IsNothing(subkey.GetValue("Version")) Then
                                dr.Component_Version = Left(Replace(subkey.GetValue("Version").ToString, ",", "."), ds.Components.Component_VersionColumn.MaxLength)
                                ' Sometimes it's a binary array
                                If dr.Component_Version = "System.Byte[]" Then
                                    Dim b() As Byte
                                    b = CType(subkey.GetValue("Version"), Byte())
                                    temp = Misc.ByteArrayToString(b)
                                    dr.Component_Version = Left(temp, ds.Components.Component_VersionColumn.MaxLength)
                                End If
                            End If
                            If Component = "" And ComponentName <> "" Then
                                dr.Component = ComponentName
                            End If
                            ds.Components.AddComponentsRow(dr)
                            cnt += 1
                        End If
                    End If
                    subkey.Close()
                Next
                key.Close()
            End If

            ' Manually add an entry for MDAC
            temp = Misc.ConvUNC(RemotePC, CommonProgramFiles & "\System\msadc\msadco.dll")
            If File.Exists(temp) Then
                dr = ds.Components.NewComponentsRow
                dr.ID_Component = ID
                dr.Component_ID = cnt
                dr.Component = "Microsoft Data Access Components"
                dr.Component_Name = "MDAC"
                dr.Component_Version = Left(Split(Misc.GetVersion(temp), " ")(0), ds.Components.Component_VersionColumn.MaxLength)
                ds.Components.AddComponentsRow(dr)
            End If

            ' Sort the list to make it look pretty
            Misc.DataTableSort(ds.Tables("Components"), "Component")

            ' now renumber the Compoent_ID field to match the sort
            cnt = 0
            For Each dr In ds.Components
                dr.Component_ID = cnt
                cnt += 1
            Next
        Catch ex As Exception
            dr = ds.Components.NewComponentsRow
            dr.ID_Component = ID
            dr.Component_ID = -1
            dr.Component = Left("Error: " & ex.Message, ds.Components.ComponentColumn.MaxLength)
            ds.Components.AddComponentsRow(dr)
            Misc.ErrorLog(ex, MachineName, "Components", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the Computer's CPU data
    '
    Public Sub GetCPU()
        Dim obj As ManagementObject
        Dim dr As SOSOSDataset.CPURow
        Dim cnt, num As Integer
        Dim temp As String

        Try
            cnt = 0
            dr = ds.CPU.NewCPURow
            dr.ID_CPU = ID

            ' This is a bit different... we don't want multiple entries for PCs
            ' with dual processsors, instead we collapse the data into a single
            ' data row and add a CPU count.
            wmi.Path.RelativePath = "Win32_Processor"
            For Each obj In wmi.GetInstances()
                ' We do not count the hyperthreading CPUs as real CPUs
                ' BTW: I've found a few systems that can fool this simple test
                If cnt > 0 And Not IsNothing(obj("ProcessorID")) AndAlso obj("ProcessorID").ToString = "0000000000000000" Then
                    Exit For
                End If
                cnt += 1
                ' we prefer Name, but Caption will do in a pinch
                If IsNothing(obj("Name")) Then
                    dr.CPU = Left(obj("Caption").ToString.Trim, ds.CPU.CPUColumn.MaxLength)
                    dr.CPU_Name = Left(obj("Caption").ToString.Trim, ds.CPU.CPU_NameColumn.MaxLength)
                Else
                    dr.CPU = Left(obj("Name").ToString.Trim, ds.CPU.CPUColumn.MaxLength)
                    dr.CPU_Name = Left(obj("Name").ToString.Trim, ds.CPU.CPU_NameColumn.MaxLength)
                End If
                If Not IsNothing(obj("SocketDesignation")) Then
                    dr.CPU_Socket = Left(obj("SocketDesignation").ToString, ds.CPU.CPU_SocketColumn.MaxLength)
                End If
                num = Convert.ToInt32(obj("CurrentClockSpeed"))
                ' for PC's on battery power
                If Not IsNothing(obj("MaxClockSpeed")) Then
                    num = Convert.ToInt32(obj("MaxClockSpeed"))
                End If
                dr.CPU_ExtSpeed = 0
                If Not IsNothing(obj("ExtClock")) Then
                    dr.CPU_ExtSpeed = Convert.ToInt32(obj("ExtClock"))
                End If
                dr.CPU_Speed = Misc.RoundCPUSpeed(num, dr.CPU_ExtSpeed)
                dr.CPU_Count = cnt
                ' The value for CPU_Name is often too complex to sort on, so
                ' we generate a simpler name here
                If Not IsNothing(obj("Manufacturer")) And Not IsNothing(obj("ProcessorID")) AndAlso obj("ProcessorID").ToString.Length = 16 Then
                    dr.CPU = Misc.GetCPUType(obj("Manufacturer").ToString, obj("ProcessorID").ToString)
                End If
            Next

            ' Check the newer WMI fields for the number of CPU cores
            ' This works in WinXP SP3, Vista, WinS2008, and WinS2003 with HotFix KB932370
            Dim NumberOfCores, NumberOfProcessors As Integer
            NumberOfCores = 0
            NumberOfProcessors = 0
            For Each obj In wmi.GetInstances()
                For Each prop As PropertyData In obj.Properties
                    If prop.Name = "NumberOfCores" Then
                        NumberOfCores = Convert.ToInt32(obj("NumberOfCores"))
                        Exit For
                    End If
                Next
                Exit For
            Next
            wmi.Path.RelativePath = "Win32_ComputerSystem"
            For Each obj In wmi.GetInstances()
                NumberOfProcessors = Convert.ToInt32(obj("NumberOfProcessors"))
            Next
            ' if NumberOfCores is availabe, then NumberOfProcessors contains only
            ' physical CPU chips, otherwise it contains hyperthreading in its count
            If NumberOfCores > 0 And NumberOfProcessors > 0 Then
                dr.CPU_Count = NumberOfCores * NumberOfProcessors
            End If

            ' shorten it up a bit for the Summary SOS table
            temp = dr.CPU
            If Not temp.IndexOf("Family") >= 0 Then
                temp = Replace(temp, "Intel ", "")
            End If
            temp = Replace(temp, " processor", "")
            ds.SOS(0).CPU_Type = temp
            ds.SOS(0).CPU_Speed = dr.CPU_Speed
            ds.SOS(0).CPU_Count = dr.CPU_Count

            ds.CPU.AddCPURow(dr)
        Catch ex As Exception
            dr = ds.CPU.NewCPURow
            dr.ID_CPU = ID
            dr.CPU = Left("Error: " & ex.Message, ds.CPU.CPUColumn.MaxLength)
            ds.CPU.AddCPURow(dr)
            Misc.ErrorLog(ex, MachineName, "CPU", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the user's Desktop settings
    '
    Public Sub GetDesktop()
        If ProfileList.Count > 0 Then
            Dim ProfileName, Profile, temp, MountPoint As String
            Dim key As RegistryKey
            Dim dr As SOSOSDataset.DesktopRow

            ProfileName = ""
            Try
                ' look at each user's loaded registry hive
                For Each Profile In ProfileList.Keys
                    ProfileName = Left(Profile, ds.Desktop.ID_ProfileColumn.MaxLength)
                    MountPoint = ProfileList(Profile).TrimStart("*"c)

                    key = reg_hku.OpenSubKey(MountPoint & "\Control Panel\Desktop", False)
                    If Not IsNothing(key) Then
                        dr = ds.Desktop.NewDesktopRow
                        dr.ID_Desktop = ID
                        dr.ID_Profile = ProfileName
                        temp = key.GetValue("Wallpaper", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString
                        dr.Desktop_Wallpaper = Left(Misc.ExpandEnv(RemotePC, temp, SystemPathVars), ds.Desktop.Desktop_WallpaperColumn.MaxLength)
                        temp = key.GetValue("SCRNSAVE.EXE", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString
                        dr.Desktop_ScreenSaver = Left(Misc.ExpandEnv(RemotePC, temp, SystemPathVars), ds.Desktop.Desktop_ScreenSaverColumn.MaxLength)
                        dr.Desktop_SS_Active = CBool(key.GetValue("ScreenSaveActive", "false"))
                        dr.Desktop_SS_Timeout = CInt(key.GetValue("ScreenSaveTimeOut", "0"))
                        dr.Desktop_SS_IsSecure = CBool(key.GetValue("ScreenSaverIsSecure", "false"))
                        ds.Desktop.AddDesktopRow(dr)
                        key.Close()
                    End If
                Next
            Catch ex As Exception
                dr = ds.Desktop.NewDesktopRow
                dr.ID_Desktop = ID
                ' it's possible that an early error wouldn't populate the
                ' ProfileName field (which is a part of the foreign key)
                If ProfileName = "" Then
                    ProfileName = ds.Profiles(0).Profile
                End If
                dr.ID_Profile = ProfileName
                dr.Desktop_SS_Active = False
                dr.Desktop_SS_IsSecure = False
                dr.Desktop_Wallpaper = Left("Error: " & ex.Message, ds.Desktop.Desktop_WallpaperColumn.MaxLength)
                ds.Desktop.AddDesktopRow(dr)
                Misc.ErrorLog(ex, MachineName, "Desktop", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the list of devices... similar to the listing in the Device Manager
    '
    Public Sub GetDevices()
        Dim obj, related As ManagementObject
        Dim key As RegistryKey
        Dim cnt, error_row As Integer
        Dim DeviceName, CategoryName, ClassGUID As String
        Dim dr As SOSOSDataset.DevicesRow

        Try
            ' start with a high number to facilitate sorting
            cnt = 10000
            error_row = -1
            wmi.Path.RelativePath = "Win32_PnPEntity"
            For Each obj In wmi.GetInstances()
                ClassGUID = ""
                If Not IsNothing(obj("ClassGUID")) Then
                    ClassGUID = obj("ClassGUID").ToString
                End If

                ' skip the "Non-Plug and Play Drivers"
                If ClassGUID <> "{8ECC055D-047F-11D1-A537-0000F8753ED1}" Then
                    dr = ds.Devices.NewDevicesRow
                    dr.ID_Device = ID
                    dr.Device_ID = cnt
                    dr.Device_PnPDeviceID = Left(obj("DeviceID").ToString, ds.Devices.Device_PnPDeviceIDColumn.MaxLength)

                    ' rare case where Name isn't populated
                    If IsNothing(obj("Name")) Then
                        DeviceName = obj("DeviceID").ToString
                    Else
                        DeviceName = obj("Name").ToString
                    End If
                    dr.Device = Left(DeviceName, ds.Devices.DeviceColumn.MaxLength)

                    ' Translate the ClassGUID into a user-friendly Category
                    CategoryName = ""
                    If ClassGUID <> "" Then
                        If IsWin32 Then
                            key = reg_hklm.OpenSubKey("SYSTEM\CurrentControlSet\Control\Class\" & ClassGUID, False)
                        Else
                            key = reg_hklm.OpenSubKey("SYSTEM\CurrentControlSet\Services\Class\" & ClassGUID, False)
                        End If
                        If Not IsNothing(key) Then
                            CategoryName = CStr(key.GetValue("", ""))
                            If CategoryName = "" Or CategoryName.StartsWith("@") Then
                                CategoryName = CStr(key.GetValue("Class", ""))
                            End If
                            key.Close()
                        End If
                    End If
                    If CategoryName = "" Then
                        CategoryName = "Other"
                    End If
                    dr.Device_Category = Left(CategoryName, ds.Devices.Device_CategoryColumn.MaxLength)

                    ' look up the PathName in the related class
                    ' Win32_SystemDriver (this takes a long time)
                    Try
                        For Each related In obj.GetRelated("Win32_SystemDriver")
                            If Not IsNothing(related("PathName")) Then
                                dr.Device_Driver = Left(related("PathName").ToString, ds.Devices.Device_DriverColumn.MaxLength)
                            End If
                        Next
                    Catch mex As System.Management.ManagementException When mex.ErrorCode = ManagementStatus.NotFound
                        ' ignore
                    End Try

                    If Not IsNothing(obj("Status")) AndAlso obj("Status").ToString <> "" Then
                        dr.Device_Status = Left(obj("Status").ToString, ds.Devices.Device_StatusColumn.MaxLength)
                    End If

                    ds.Devices.AddDevicesRow(dr)
                    cnt += 1
                End If
            Next

            ' Sort the datarows
            Misc.DataTableSort(ds.Devices, "Device_Category, Device")

            ' Now redo the device ID values to match the new sort
            cnt = 0
            For Each dr In ds.Devices
                If dr.Device_ID > 0 Then
                    dr.Device_ID = cnt
                    cnt += 1

                    ' see if we've got a SmartCard reader
                    If dr.Device.ToUpper Like "*SMARTCARD*" Or dr.Device.ToUpper Like "*SMART CARD*" Then
                        ds.SOS(0).SmartCard = True
                    End If
                    ' ok, how about a SCSI card?
                    If dr.Device.ToUpper Like "*SCSI*" Then
                        ds.SOS(0).SCSI = True
                    End If
                End If
            Next

            ' a quick way to see if a sound card exists for the SOS table
            wmi.Path.RelativePath = "Win32_SoundDevice"
            For Each obj In wmi.GetInstances()
                ds.SOS(0).SoundCard = True
                Exit For
            Next
        Catch ex As Exception
            dr = ds.Devices.NewDevicesRow
            dr.ID_Device = ID
            dr.Device_ID = error_row
            dr.Device_Category = Left("Error: " & ex.Message, ds.Devices.Device_CategoryColumn.MaxLength)
            ds.Devices.AddDevicesRow(dr)
            Misc.ErrorLog(ex, MachineName, "Devices", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the physical disks
    ' Note: You should call GetDevices prior to calling this function
    '
    Public Sub GetDisks()
        Dim obj, related As ManagementObject
        Dim cnt, num As Integer
        Dim temp As String
        Dim dr As SOSOSDataset.DisksRow

        Try
            cnt = 0
            wmi.Path.RelativePath = "Win32_DiskDrive"
            For Each obj In wmi.GetInstances()
                dr = ds.Disks.NewDisksRow
                dr.ID_Disk = ID
                dr.Disk_ID = cnt
                If Not IsNothing(obj("Manufacturer")) Then
                    dr.Disk_Desc = Left(obj("Manufacturer").ToString, ds.Disks.Disk_DescColumn.MaxLength)
                End If
                If Not IsNothing(obj("Model")) Then
                    dr.Disk_Desc = Left(obj("Model").ToString, ds.Disks.Disk_DescColumn.MaxLength)
                End If
                If dr.IsDisk_DescNull Then
                    dr.Disk_Desc = Left(obj("Name").ToString, ds.Disks.Disk_DescColumn.MaxLength)
                End If
                If Not IsNothing(obj("InterfaceType")) Then
                    dr.Disk_Interface = Left(obj("InterfaceType").ToString, ds.Disks.Disk_InterfaceColumn.MaxLength)
                End If
                If Not IsNothing(obj("Partitions")) Then
                    dr.Disk_Partitions = Convert.ToInt32(obj("Partitions"))
                End If
                ' we use the Plug and Play device ID as the "starter value"
                If Not IsNothing(obj("PnPDeviceID")) Then
                    dr.Disk_SerialNumber = Left(Path.GetFileName(obj("PnPDeviceID").ToString), ds.Disks.Disk_SerialNumberColumn.MaxLength)
                End If
                ' let's see if we can get a better Serial Number
                For Each related In obj.GetRelated("Win32_PhysicalMedia")
                    Dim i As Integer
                    Dim c As Char
                    Dim IsRaw As Boolean
                    Dim ValidHex As String = "0123456789abcdefABCDEF"

                    If Not IsNothing(related("SerialNumber")) Then
                        temp = related("SerialNumber").ToString
                        ' as usual, some sanity checking
                        If temp = "" Then
                            Exit For
                        End If

                        ' let's see if this is really an ASCII hex string
                        For Each c In temp.ToCharArray
                            If ValidHex.IndexOf(c) < 0 Then
                                IsRaw = True
                                Exit For
                            End If
                        Next

                        If Not IsRaw Then
                            ' convert a hex string into a string
                            Dim sb As New System.Text.StringBuilder

                            For i = 0 To temp.Length - 1 Step 2
                                c = Chr(Integer.Parse(temp.Substring(i, 2), NumberStyles.AllowHexSpecifier))
                                If Char.IsControl(c) Then
                                    ' yikes, this isn't a Hex string after all
                                    sb.Length = 0
                                    sb.Append(temp)
                                    Exit For
                                End If
                                sb.Append(c)
                            Next
                            dr.Disk_SerialNumber = Left(Trim(sb.ToString), ds.Disks.Disk_SerialNumberColumn.MaxLength)
                        Else
                            dr.Disk_SerialNumber = Left(Trim(temp), ds.Disks.Disk_SerialNumberColumn.MaxLength)
                        End If
                    End If
                Next
                ds.Disks.AddDisksRow(dr)
                cnt += 1
            Next

            ' next, do the CDROMs (we sometimes get a Provide Failure)
            wmi.Path.RelativePath = "Win32_CDROMDrive"
            For Each obj In wmi.GetInstances()
                dr = ds.Disks.NewDisksRow
                dr.ID_Disk = ID
                dr.Disk_ID = cnt
                If Not IsNothing(obj("Name")) Then
                    If obj("Name").ToString.Length < 3 And Not IsNothing(obj("Description")) Then
                        dr.Disk_Desc = Left(obj("Description").ToString, ds.Disks.Disk_DescColumn.MaxLength)
                    Else
                        dr.Disk_Desc = Left(obj("Name").ToString, ds.Disks.Disk_DescColumn.MaxLength)
                    End If
                Else
                    ' As a last resort, we'll take the DeviceID
                    dr.Disk_Desc = obj("DeviceID").ToString
                End If
                If Not IsNothing(obj("PnPDeviceID")) Then
                    ' the PnPDeviceID is the only "serial number" we've got
                    dr.Disk_SerialNumber = Left(Path.GetFileName(obj("PnPDeviceID").ToString), ds.Disks.Disk_SerialNumberColumn.MaxLength)
                End If

                ' Since InterfaceType isn't supported for CDROMs, we fake it.
                ' Some CDROM show up as SCSI devices in the registry, even
                ' when they are not...
                temp = obj("DeviceID").ToString
                num = InStr(temp, "\") - 1
                If num > 0 Then
                    dr.Disk_Interface = Left(temp, num)
                    If ds.SOS(0).SCSI = False And dr.Disk_Interface = "SCSI" Then
                        dr.Disk_Interface = "IDE"
                    End If
                End If

                ds.Disks.AddDisksRow(dr)
                cnt += 1
            Next
        Catch ex As Exception
            dr = ds.Disks.NewDisksRow
            dr.ID_Disk = ID
            dr.Disk_ID = -1
            dr.Disk_Desc = Left("Error: " & ex.Message, ds.Disks.Disk_DescColumn.MaxLength)
            ds.Disks.AddDisksRow(dr)

            ' Check to see if this is an error that is fairly common.  If so,
            ' let's not clog up the error log with it.
            If ex.GetType Is GetType(System.Management.ManagementException) Then
                If CType(ex, ManagementException).ErrorCode = ManagementStatus.ProviderFailure Then
                    Misc.ErrorLog(ex, MachineName, "Disks", Misc.ErrLogLevel.Full)
                    Exit Try
                End If
            End If
            Misc.ErrorLog(ex, MachineName, "Disks", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    ' 
    ' Get the information about the locally attached drive letters
    '
    Public Sub GetDrives()
        Dim obj As ManagementObject
        Dim dr As SOSOSDataset.DrivesRow

        Try
            ds.SOS(0).Disk_Size = 0
            ds.SOS(0).Disk_Free = 0
            ds.SOS(0).Disk_Count = 0

            wmi.Path.RelativePath = "Win32_LogicalDisk"
            For Each obj In wmi.GetInstances()
                ' skip network drives for now (we catch them later)
                If Convert.ToInt32(obj("drivetype")) <> 4 Then
                    dr = ds.Drives.NewDrivesRow
                    dr.ID_Drive = ID
                    dr.Drive = Left(obj("DeviceID").ToString, ds.Drives.DriveColumn.MaxLength)
                    If Not IsNothing(obj("Size")) AndAlso obj("size").ToString.Length < 20 Then
                        dr.Drive_Size = CInt(Convert.ToDouble(obj("Size")) / (1024.0# * 1024.0#))
                    Else
                        dr.Drive_Size = 0
                    End If
                    If Not IsNothing(obj("FreeSpace")) Then
                        dr.Drive_FreeSpace = CInt(Convert.ToDouble(obj("FreeSpace")) / (1024.0# * 1024.0#))
                    Else
                        dr.Drive_FreeSpace = 0
                    End If
                    dr.Drive_Type = Choose(Convert.ToInt32(obj("DriveType")) + 1, "Unknown", "NoRoot", "Removable", "Local", "Network", "CDROM", "RAM").ToString
                    If Not IsNothing(obj("FileSystem")) Then
                        dr.Drive_FileSystem = Left(obj("FileSystem").ToString, ds.Drives.Drive_FileSystemColumn.MaxLength)
                    End If
                    If Not IsNothing(obj("MediaType")) Then
                        dr.Drive_MediaType = Choose(Convert.ToInt32(obj("mediatype")) + 1, "Unknown", "5.25 1.2m", "3.5 1.44m", "3.5 2.88m", "3.5 20.8m", "3.5 720k", "5.25 360k", "5.25 320k", "5.25 320k", "5.25 180k", "5.25 160k", "Other", "Fixed", "LS120", "3.5 640k", "5.25 640k", "5.25 720k", "3.5 1.2m", "3.5 1.23m", "5.25 1.23m", "3.25 128m", "3.25 230m", "8 256k").ToString
                    End If
                    If Not IsNothing(obj("VolumeName")) Then
                        dr.Drive_VolumeName = Left(obj("VolumeName").ToString, ds.Drives.Drive_VolumeNameColumn.MaxLength)
                    End If
                    ' combine all of the fixed disk's numbers together for the
                    ' SOS Summary table
                    If Convert.ToInt32(obj("drivetype")) = 3 Then
                        ds.SOS(0).Disk_Size += dr.Drive_Size
                        ds.SOS(0).Disk_Free += dr.Drive_FreeSpace
                        ds.SOS(0).Disk_Count += 1
                    End If
                    If Convert.ToInt32(obj("drivetype")) = 5 Then
                        ds.SOS(0).CDROM = True
                    End If
                    ds.Drives.AddDrivesRow(dr)
                End If
            Next
        Catch ex As Exception
            dr = ds.Drives.NewDrivesRow
            dr.ID_Drive = ID
            dr.Drive = Left("Error: " & ex.Message, ds.Drives.DriveColumn.MaxLength)
            ds.Drives.AddDrivesRow(dr)
            Misc.ErrorLog(ex, MachineName, "Drives", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the user's Email settings
    ' Note: We do NOT look at the user's email, we merely record some of the
    ' settings for the email application
    '
    Public Sub GetEmail()
        If ProfileList.Count > 0 Then
            Dim Profile, ProfileName, MountPoint, k, subkeyname, temp, pst_path, Email_ProfileName, ExchangeServer As String
            Dim key, subkey, sub2 As RegistryKey
            Dim cnt As Integer
            Dim pst_size As Double
            Dim pst_byte(), cached_byte() As Byte
            Dim fi As FileInfo
            Dim CachedMode As Boolean
            Dim dr As SOSOSDataset.EmailRow

            ProfileName = ""
            Try
                ' look at each user's loaded registry hive
                For Each Profile In ProfileList.Keys
                    ProfileName = Left(Profile, ds.Email.ID_ProfileColumn.MaxLength)
                    MountPoint = ProfileList(Profile).TrimStart("*"c)
                    cnt = 0

                    key = reg_hku.OpenSubKey(MountPoint & "\Software\Microsoft\Windows NT\CurrentVersion\Windows Messaging Subsystem\Profiles", False)
                    If IsNothing(key) Then
                        key = reg_hku.OpenSubKey(MountPoint & "\Software\Microsoft\Office\15.0\Outlook\Profiles", False)
                    End If
                    If Not IsNothing(key) Then
                        For Each k In key.GetSubKeyNames()
                            Email_ProfileName = k

                            ' Get the generic information about the Email Profile
                            CachedMode = False
                            ExchangeServer = ""
                            subkey = key.OpenSubKey(k & "\13dbb0c8aa05101a9bb000aa002fc45a", False)
                            If Not IsNothing(subkey) Then

                                ' get the "cached mode" flag
                                cached_byte = CType(subkey.GetValue("00036601"), Byte())
                                If Not IsNothing(cached_byte) Then
                                    If cached_byte(0) = &H84 Then
                                        CachedMode = True
                                    End If
                                End If

                                ' get the Exchange server path
                                ExchangeServer = subkey.GetValue("001e6608", "").ToString
                                subkey.Close()
                            End If

                            ' search for PST files
                            pst_path = ""
                            subkey = key.OpenSubKey(k, False)
                            For Each subkeyname In subkey.GetSubKeyNames()
                                sub2 = subkey.OpenSubKey(subkeyname, False)

                                ' Note: Outlook does a poor job of cleaning up after itself, so it's
                                ' possible to get information on PST files that are no longer in use.

                                ' PST for versions of Outlook prior to 2003
                                temp = sub2.GetValue("001e6700", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString
                                temp = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
                                If temp <> "" Then
                                    pst_path = temp
                                    pst_size = 0
                                    ' We may get a permission denied (particularly in
                                    ' a workgroup environment)
                                    If File.Exists(Misc.ConvUNC(RemotePC, pst_path)) Then
                                        fi = New FileInfo(Misc.ConvUNC(RemotePC, pst_path))
                                        pst_size = fi.Length / 1024.0
                                    End If
                                    dr = ds.Email.NewEmailRow
                                    dr.ID_Email = ID
                                    dr.ID_Profile = ProfileName
                                    dr.Email_ProfileName = Left(Email_ProfileName, ds.Email.Email_ProfileNameColumn.MaxLength)
                                    dr.Email_ID = cnt
                                    dr.Email_CachedMode = CachedMode
                                    dr.Email_PST_Path = Left(pst_path, ds.Email.Email_PST_PathColumn.MaxLength)
                                    dr.Email_PST_Size = pst_size
                                    If ExchangeServer <> "" Then
                                        dr.Email_ExchangeServer = Left(ExchangeServer, ds.Email.Email_ExchangeServerColumn.MaxLength)
                                    End If
                                    ds.Email.AddEmailRow(dr)
                                    cnt += 1
                                End If

                                ' PST for Outlook 2003 and beyond
                                pst_byte = CType(sub2.GetValue("001f6700"), Byte())
                                If Not IsNothing(pst_byte) Then
                                    pst_path = System.Text.Encoding.Unicode.GetString(pst_byte, 0, UBound(pst_byte) - 1)
                                    pst_size = 0
                                    If File.Exists(Misc.ConvUNC(RemotePC, pst_path)) Then
                                        fi = New FileInfo(Misc.ConvUNC(RemotePC, pst_path))
                                        pst_size = fi.Length / 1024.0
                                    End If
                                    dr = ds.Email.NewEmailRow
                                    dr.ID_Email = ID
                                    dr.ID_Profile = ProfileName
                                    dr.Email_ProfileName = Left(Email_ProfileName, ds.Email.Email_ProfileNameColumn.MaxLength)
                                    dr.Email_ID = cnt
                                    dr.Email_CachedMode = CachedMode
                                    dr.Email_PST_Path = Left(pst_path, ds.Email.Email_PST_PathColumn.MaxLength)
                                    dr.Email_PST_Size = pst_size
                                    If ExchangeServer <> "" Then
                                        dr.Email_ExchangeServer = Left(ExchangeServer, ds.Email.Email_ExchangeServerColumn.MaxLength)
                                    End If
                                    ds.Email.AddEmailRow(dr)
                                    cnt += 1
                                End If
                                sub2.Close()
                            Next

                            ' Even if we don't have a PST, we still want to
                            ' record the other settings
                            If pst_path = "" Then
                                dr = ds.Email.NewEmailRow
                                dr.ID_Email = ID
                                dr.ID_Profile = ProfileName
                                dr.Email_ProfileName = Left(Email_ProfileName, ds.Email.Email_ProfileNameColumn.MaxLength)
                                dr.Email_ID = cnt
                                dr.Email_CachedMode = CachedMode
                                dr.Email_PST_Size = 0
                                If ExchangeServer <> "" Then
                                    dr.Email_ExchangeServer = Left(ExchangeServer, ds.Email.Email_ExchangeServerColumn.MaxLength)
                                End If
                                ds.Email.AddEmailRow(dr)
                                cnt += 1
                            End If
                            subkey.Close()
                        Next
                        key.Close()
                    End If
                Next
            Catch ex As Exception
                dr = ds.Email.NewEmailRow
                dr.ID_Email = ID
                ' it's possible that an early error wouldn't populate the
                ' ProfileName field (which is a part of the foreign key)
                If ProfileName = "" Then
                    ProfileName = ds.Profiles(0).Profile
                End If
                dr.ID_Profile = ProfileName
                dr.Email_ID = -1
                dr.Email_CachedMode = False
                dr.Email_ProfileName = Left("Error: " & ex.Message, ds.Email.Email_ProfileNameColumn.MaxLength)
                ds.Email.AddEmailRow(dr)
                Misc.ErrorLog(ex, MachineName, "Email", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get information about the Event Log Settings
    ' Note: Should be run before GetEventlogs() below
    '
    Public Sub GetEventLogSettings()
        If IsWin32 Then
            Dim dr As SOSOSDataset.EventLogSettingsRow
            Dim i As Integer
            Dim obj As ManagementObject

            Try
                ' Hummm... this doesn't detect all custom eventlogs
                wmi.Path.RelativePath = "Win32_NTEventlogFile"
                For Each obj In wmi.GetInstances()
                    dr = ds.EventLogSettings.NewEventLogSettingsRow
                    dr.ID_EventSettings = ID
                    dr.EventSettings_Logfile = Left(obj("LogfileName").ToString, ds.EventLogSettings.EventSettings_LogfileColumn.MaxLength)
                    dr.EventSettings_CurrentSize = CInt(Convert.ToInt64(obj("FileSize")) / 1024)
                    dr.EventSettings_MaxSize = CInt(Convert.ToInt32(obj("MaxFileSize")) / 1024)
                    If Not IsNothing(obj("NumberOfRecords")) Then
                        dr.EventSettings_NumEntries = Convert.ToInt32(obj("NumberOfRecords"))
                    End If
                    i = Convert.ToInt32(obj("OverwriteOutDated"))
                    If i < 0 Or i > 365 Then
                        ' indicates an overwrite policy of "none"
                        i = -1
                    End If
                    dr.EventSettings_Retension = i
                    ds.EventLogSettings.AddEventLogSettingsRow(dr)
                Next
            Catch ex As Exception
                dr = ds.EventLogSettings.NewEventLogSettingsRow
                dr.ID_EventSettings = ID
                dr.EventSettings_Logfile = Left("Error: " & ex.Message, ds.EventLogSettings.EventSettings_LogfileColumn.MaxLength)
                ds.EventLogSettings.AddEventLogSettingsRow(dr)
                Misc.ErrorLog(ex, MachineName, "EventLogSettings", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the EventLog data
    ' Note: Administrator rights are required to retrieve data from the
    ' Security Log
    '
    Public Sub GetEventLogs()
        If IsWin32 Then
            Dim el As EventLog
            Dim els() As EventLog
            Dim eles(0), ele As EventLogEntry
            Dim date_marker As Date
            Dim temp As String
            Dim day_limit, row_limit, min_row, error_row, cnt, i As Integer
            Dim dr As SOSOSDataset.EventLogsRow
            Dim dr_settings As SOSOSDataset.EventLogSettingsRow

            Try
                error_row = -1
                ' Return data for just the number of days requested and/or the
                ' total number of rows per logfile 
                day_limit = CInt(My.Settings.EventsByDay)
                row_limit = CInt(My.Settings.EventsByLine)

                date_marker = Date.MinValue
                If day_limit > 0 Then
                    date_marker = DateAdd(DateInterval.Day, -day_limit, Now())
                End If

                ' This is the most time-consuming part of the entire
                ' application.  In my case, we collect 7 days of eventlogs.
                ' For PCs that's no big deal... about 2500 entries and takes
                ' a few seconds.  But on servers, that equates to 50,000 
                ' entries and takes 1-2 minutes.

                If RemotePC = "" Then
                    els = EventLog.GetEventLogs()
                Else
                    els = EventLog.GetEventLogs(RemotePC)
                End If
                For Each el In els
                    Try
                        If el.Entries.Count > 0 Then
                            ' Start by copying the entries.  We pad the array
                            ' a wee bit since things tend to change rather
                            ' rapidly when dealing with the eventlog.  This
                            ' may fail the first time it's tried, so we try it
                            ' again up to 5 times...

                            For i = 0 To 4
                                Try
                                    ReDim eles(el.Entries.Count + 10)
                                    Array.Clear(eles, 0, eles.Length)
                                    el.Entries.CopyTo(eles, 0)
                                    Exit For
                                Catch ioe_ex As System.InvalidOperationException
                                    ' OK, OK... I give up!
                                    If i = 4 Then
                                        ' probably a corrupted event log
                                        Throw
                                    End If
                                End Try
                            Next

                            If row_limit > 0 And row_limit < eles.Length Then
                                min_row = eles.Length - row_limit
                            Else
                                min_row = 0
                            End If

                            ' Fill in the missing data for the OldestEntry
                            ' column in the EventLogSettings table
                            dr_settings = ds.EventLogSettings.FindByID_EventSettingsEventSettings_Logfile(ID, el.Log)
                            If Not IsNothing(dr_settings) And Not IsNothing(eles(0)) Then
                                dr_settings.EventSettings_OldestEntry = eles(0).TimeGenerated
                            End If

                            ' process in reverse cronlogical order (just like
                            ' the event viewer applet), newest at the top
                            cnt = 0
                            For i = eles.Length - 1 To min_row Step -1
                                ele = eles(i)
                                If Not IsNothing(ele) Then
                                    If day_limit > 0 And ele.TimeGenerated < date_marker Then
                                        Exit For
                                    End If

#If TRIM_SUCCESS_PRIVILEGE_USE Then
                                    If ele.EntryType <> EventLogEntryType.SuccessAudit Or ele.CategoryNumber <> 4 Then
#Else
                                    If True Then
#End If
                                        dr = ds.EventLogs.NewEventLogsRow
                                        dr.ID_Event = ID
                                        dr.Event_ID = cnt
                                        dr.Event_Logfile = Left(el.Log, ds.EventLogs.Event_LogfileColumn.MaxLength)
                                        dr.Event_Type = Left(ele.EntryType.ToString, ds.EventLogs.Event_TypeColumn.MaxLength)
                                        dr.Event_Date = ele.TimeGenerated
                                        dr.Event_Source = Left(ele.Source, ds.EventLogs.Event_SourceColumn.MaxLength)
                                        dr.Event_Code = CInt(ele.InstanceId And &HFFFF)
                                        If Not IsNothing(ele.UserName) Then
                                            dr.Event_User = Left(ele.UserName, ds.EventLogs.Event_UserColumn.MaxLength)
                                        End If
                                        If Not IsNothing(ele.Message) Then
                                            ' Clean up the message a wee bit
                                            temp = ele.Message
                                            temp = temp.Replace(vbTab, " ")
                                            temp = temp.Replace(vbCrLf, " ")
                                            temp = temp.Replace(vbCr, " ")
                                            temp = temp.Replace(vbLf, " ")
                                            temp = temp.Replace("  ", " ")
                                            dr.Event_Message = Left(temp.Trim, ds.EventLogs.Event_MessageColumn.MaxLength)
                                        End If
                                        ds.EventLogs.AddEventLogsRow(dr)
                                        cnt += 1
                                    End If
                                End If
                            Next
                        End If
                    Catch ex As Exception
                        ' Non-fatal error... Example: Non-administrators can't 
                        ' view the contents of the Security log, so we catch
                        ' this error and continue
                        dr = ds.EventLogs.NewEventLogsRow
                        dr.ID_Event = ID
                        dr.Event_ID = error_row
                        If Not IsNothing(el) AndAlso el.Log <> "" Then
                            dr.Event_Logfile = Left(el.Log, ds.EventLogs.Event_LogfileColumn.MaxLength)
                        Else
                            dr.Event_Logfile = "Unknown"
                        End If
                        dr.Event_Source = Left("Error: " & ex.Message, ds.EventLogs.Event_SourceColumn.MaxLength)
                        ds.EventLogs.AddEventLogsRow(dr)
                        error_row -= 1

                        ' Record this only if it's an "unexpected" error
                        If Not ex.GetType Is GetType(System.Security.SecurityException) Then
                            Misc.ErrorLog(ex, MachineName, "EventLogs", Misc.ErrLogLevel.Errors)
                        End If
                    End Try
                Next
            Catch ex As Exception
                ' generic fatal errors
                dr = ds.EventLogs.NewEventLogsRow
                dr.ID_Event = ID
                dr.Event_ID = error_row
                dr.Event_Logfile = "Unknown"
                dr.Event_Source = Left("Error: " & ex.Message, ds.EventLogs.Event_SourceColumn.MaxLength)
                ds.EventLogs.AddEventLogsRow(dr)
                Misc.ErrorLog(ex, MachineName, "EventLogs", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get information about an ad-hoc lists of files
    ' Note: duplicate entries in the FileInfoList create exceptions
    '
    Public Sub GetFileInfo()
        Dim fvi As FileVersionInfo
        Dim fi As FileInfo
        Dim rawfile, expfile, uncfile, filename, version As String
        Dim dr As SOSOSDataSet.FileInfoRow

        Try
            For Each rawfile In My.Settings.FileInfoList
                If rawfile <> "" Then
                    ' in case you didn't read the documentation :)
                    If rawfile.ToUpper.StartsWith("\") Then
                        rawfile = SystemDrive & rawfile
                    End If
                    expfile = Misc.ExpandEnv(RemotePC, rawfile, SystemPathVars)
                    uncfile = Misc.ConvUNC(RemotePC, expfile)
                    ' check to see if the directory exists first, then expand 
                    ' any filename wildcards
                    If Directory.Exists(Path.GetDirectoryName(uncfile)) Then
                        For Each filename In Directory.GetFiles(Path.GetDirectoryName(uncfile), Path.GetFileName(uncfile))
                            If File.Exists(filename) Then
                                fvi = FileVersionInfo.GetVersionInfo(filename)
                                ' sometimes the FileVersion is empty...
                                If fvi.FileVersion <> "" Then
                                    version = fvi.FileVersion.Replace(",", ".")
                                Else
                                    version = fvi.FileMajorPart & "." & fvi.FileMinorPart & "." & fvi.FileBuildPart
                                End If
                                fi = New FileInfo(filename)
                                dr = ds.FileInfo.NewFileInfoRow
                                dr.ID_FileInfo = ID
                                dr.FileInfo_Path = Left(Path.Combine(Path.GetDirectoryName(expfile), Path.GetFileName(filename)), ds.FileInfo.FileInfo_PathColumn.MaxLength)
                                dr.FileInfo_Version = Left(version, ds.FileInfo.FileInfo_VersionColumn.MaxLength)
                                dr.FileInfo_Size = CInt(fi.Length)
                                dr.FileInfo_ModDate = fi.LastWriteTime
                                ' It's possible to have a duplicate if you included both the
                                ' 32 and 64 bit versions of the environmental variables
                                If IsNothing(ds.FileInfo.FindByID_FileInfoFileInfo_Path(ID, dr.FileInfo_Path)) Then
                                    ds.FileInfo.AddFileInfoRow(dr)
                                End If
                            End If
                        Next
                    End If
                End If
            Next
        Catch ex As Exception
            dr = ds.FileInfo.NewFileInfoRow
            dr.ID_FileInfo = ID
            dr.FileInfo_Path = Left("Error: " & ex.Message, ds.FileInfo.FileInfo_PathColumn.MaxLength)
            ds.FileInfo.AddFileInfoRow(dr)
            Misc.ErrorLog(ex, MachineName, "FileInfo", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Install Version.  We use a text file on each PC to record the
    ' installation version... typically formatted with a description and date
    ' separated by a tab.  It looks like this:
    ' v1.0.0	17 Mar 03
    ' Note: See the documentation for additional information
    '
    Public Sub GetInstallVer()
        Dim filename, rawfile, temp As String
        Dim sr As StreamReader

        ' This routine writes to the SOS Summary table, there is no
        ' separate "InstallVer" table.
        Try
            rawfile = My.Settings.VersionFile
            If rawfile <> "" Then
                ' in case you didn't read the documentation :)
                If rawfile.StartsWith("\") Then
                    rawfile = SystemDrive & rawfile
                End If
                rawfile = Misc.ExpandEnv(RemotePC, rawfile, SystemPathVars)
                filename = Misc.ConvUNC(RemotePC, rawfile)
                If File.Exists(filename) Then
                    sr = [File].OpenText(filename)
                    Do While sr.Peek() > -1
                        temp = sr.ReadLine
                        ' if there are multiple entries, just keep the last one
                        If temp <> "" And Not temp.StartsWith(" ") And Not temp.StartsWith(ControlChars.Tab) Then
                            If temp.StartsWith("V") Then
                                temp = "v" & temp.Substring(1)
                            End If
                            ds.SOS(0).Install_Rev = Left(Replace(temp, ControlChars.Tab, " - "), ds.SOS.Install_RevColumn.MaxLength)
                        End If
                    Loop
                    sr.Close()
                End If
            End If

            ' Install Date for the OS
            wmi.Path.RelativePath = "Win32_OperatingSystem"
            For Each obj As ManagementObject In wmi.GetInstances()
                If Not IsNothing(obj("InstallDate")) Then
                    ds.SOS(0).Install_Date = ManagementDateTimeConverter.ToDateTime(obj("InstallDate").ToString)
                End If
            Next
            If IsNothing(ds.SOS(0).Install_Date) AndAlso Directory.Exists(Misc.ConvUNC(RemotePC, SystemRoot)) Then
                ds.SOS(0).Install_Date = File.GetCreationTime(Misc.ConvUNC(RemotePC, SystemRoot))
                ' on a FAT parition, this returns zero FileTime structure
                If ds.SOS(0).Install_Date < [Date].Parse("01/01/1980", culture) Then
                    ds.SOS(0).Install_Date = File.GetLastWriteTime(Misc.ConvUNC(RemotePC, SystemRoot))
                End If
            End If
        Catch ex As Exception
            ds.SOS(0).Install_Rev = Left("Error:" & ex.Message, ds.SOS.Install_RevColumn.MaxLength)
            Misc.ErrorLog(ex, MachineName, "InstallVer", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the user's Internet settings
    '
    Public Sub GetInternet()
        If ProfileList.Count > 0 Then
            Dim Profile, ProfileName, MountPoint, temp As String
            Dim key As RegistryKey
            Dim got_it As Boolean
            Dim dr As SOSOSDataset.InternetRow

            ProfileName = ""
            Try
                ' look at each user's loaded registry hive
                For Each Profile In ProfileList.Keys
                    ProfileName = Left(Profile, ds.Internet.ID_ProfileColumn.MaxLength)
                    MountPoint = ProfileList(Profile).TrimStart("*"c)

                    dr = ds.Internet.NewInternetRow
                    dr.ID_Internet = ID
                    dr.ID_Profile = ProfileName

                    got_it = False
                    ' cachesize
                    key = reg_hku.OpenSubKey(MountPoint & "\Software\Microsoft\Windows\CurrentVersion\Internet Settings\5.0\Cache\Content", False)
                    If Not IsNothing(key) Then
                        dr.Internet_CacheSize = CInt(key.GetValue("CacheLimit", "0"))
                        got_it = True
                        key.Close()
                    End If

                    dr.Internet_ProxyEnabled = False
                    dr.Internet_ProxyBypassLocal = False
                    ' proxy settings
                    key = reg_hku.OpenSubKey(MountPoint & "\Software\Microsoft\Windows\CurrentVersion\Internet Settings", False)
                    If Not IsNothing(key) Then
                        dr.Internet_ProxyEnabled = CBool(key.GetValue("ProxyEnable", 0))
                        dr.Internet_ProxyServers = Left(key.GetValue("ProxyServer", "").ToString, ds.Internet.Internet_ProxyServersColumn.MaxLength)
                        temp = key.GetValue("ProxyOverride", "").ToString
                        dr.Internet_ProxyExceptions = Left(temp.Replace("<local>", "").Trim(";"c), ds.Internet.Internet_ProxyExceptionsColumn.MaxLength)
                        dr.Internet_ProxyBypassLocal = CBool(temp.IndexOf("<local>") >= 0)
                        got_it = True
                        key.Close()
                    End If

                    If got_it Then
                        ds.Internet.AddInternetRow(dr)
                    End If
                Next
            Catch ex As Exception
                dr = ds.Internet.NewInternetRow
                dr.ID_Internet = ID
                ' it's possible that an early error wouldn't populate the
                ' ProfileName field (which is a part of the foreign key)
                If ProfileName = "" Then
                    ProfileName = ds.Profiles(0).Profile
                End If
                dr.ID_Profile = ProfileName
                dr.Internet_ProxyEnabled = False
                dr.Internet_ProxyBypassLocal = False
                dr.Internet_ProxyServers = Left("Error: " & ex.Message, ds.Internet.Internet_ProxyServersColumn.MaxLength)
                ds.Internet.AddInternetRow(dr)
                Misc.ErrorLog(ex, MachineName, "Internet", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get information about the User's mapped network resources
    '
    Public Sub GetMapped()
        If ProfileList.Count > 0 Then
            Dim i As Integer
            Dim Profile, ProfileName, MountPoint, k As String
            Dim key, subkey As RegistryKey
            Dim dr As SOSOSDataset.MappedRow

            ProfileName = ""
            Try
                ' look at each user's loaded registry hive
                For Each Profile In ProfileList.Keys
                    ProfileName = Left(Profile, ds.Mapped.ID_ProfileColumn.MaxLength)
                    MountPoint = ProfileList(Profile).TrimStart("*"c)

                    key = reg_hku.OpenSubKey(MountPoint & "\Network", False)
                    If Not IsNothing(key) Then
                        For Each k In key.GetSubKeyNames()
                            subkey = key.OpenSubKey(k, False)

                            dr = ds.Mapped.NewMappedRow
                            dr.ID_Mapped = ID
                            dr.ID_Profile = ProfileName
                            dr.Mapped = k.ToUpper
                            dr.Mapped_Path = Left(subkey.GetValue("RemotePath", "").ToString, ds.Mapped.Mapped_PathColumn.MaxLength)
                            dr.Mapped_Provider = Left(subkey.GetValue("ProviderName", "").ToString, ds.Mapped.Mapped_ProviderColumn.MaxLength)
                            i = CInt(subkey.GetValue("ConnectionType", 0))
                            dr.Mapped_Type = Choose(i, "Disk", "Printer").ToString
                            ds.Mapped.AddMappedRow(dr)
                            subkey.Close()
                        Next
                        key.Close()
                    End If
                Next
            Catch ex As Exception
                dr = ds.Mapped.NewMappedRow
                dr.ID_Mapped = ID
                ' it's possible that an early error wouldn't populate the
                ' ProfileName field (which is a part of the foreign key)
                If ProfileName = "" Then
                    ProfileName = ds.Profiles(0).Profile
                End If
                dr.ID_Profile = ProfileName
                dr.Mapped = Left("Error: " & ex.Message, ds.Mapped.MappedColumn.MaxLength)
                ds.Mapped.AddMappedRow(dr)
                Misc.ErrorLog(ex, MachineName, "Mapped", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the type and amount of installed RAM
    '
    Public Sub GetMemory()
        Dim obj As ManagementObject
        Dim num, cnt, tot As Integer
        Dim temp As String
        Dim dr As SOSOSDataset.MemoryRow

        Try
            num = 0
            cnt = 0
            temp = ""

            ' We collapse all of the entries into a single database row
            dr = ds.Memory.NewMemoryRow
            dr.ID_Memory = ID

            wmi.Path.RelativePath = "Win32_PhysicalMemory"
            For Each obj In wmi.GetInstances()
                cnt += 1
                ' sum of the memory capacity values
                num = num + CInt(Convert.ToInt64(obj("capacity")) / (1024 * 1024))
                Select Case Convert.ToInt32(obj("typedetail"))
                    Case 1
                        temp = "Reserved"
                    Case 2
                        temp = "Other"
                    Case 4
                        temp = "Unknown"
                    Case 8
                        temp = "Fast Page Mode"
                    Case 16
                        temp = "Static Column"
                    Case 32
                        temp = "Pseudo Static"
                    Case 64
                        temp = "RAMBUS"
                    Case 128
                        temp = "SDRAM"
                    Case 256
                        temp = "CMOS"
                    Case 512
                        temp = "EDO"
                    Case 1024
                        temp = "Windows DRAM"
                    Case 2048
                        temp = "Cache DRAM"
                    Case 4096
                        temp = "Non Volatile"
                End Select
                dr.Memory = num
                dr.Memory_Type = temp
                If Not IsNothing(obj("FormFactor")) Then
                    dr.Memory_Form = Choose(Convert.ToInt32(obj("FormFactor")) + 1, "Unknown", "Other", "SIP", "DIP", "ZIP", "SOJ", "Proprietry", "SIMM", "DIMM", "TSOP", "PGA", "RIMM", "SODIMM", "SRIMM", "SMD", "SSMP", "QFP", "TQFP", "SOIC", "LLC", "PLCC", "BGA", "FPBGA", "LGA").ToString
                End If
                dr.Memory_Count = cnt
                If Not IsNothing(obj("TotalWidth")) Then
                    dr.Memory_Width = Convert.ToInt32(obj("TotalWidth"))
                End If
                If Not IsNothing(obj("speed")) Then
                    dr.Memory_Speed = Convert.ToInt32(obj("speed"))
                End If
            Next

            ' round off to a multiple of 2
            ds.SOS(0).Memory = CInt(IIf((num Mod 2) <> 0, num + 1, num))

            ' Try this alternate approach as a sanity check.  In some situations, the sum
            ' of the individual Win32_PhysicalMemory instances doesn't match the total in
            ' Win32_ComputerSystem.
            wmi.Path.RelativePath = "Win32_ComputerSystem"
            For Each obj In wmi.GetInstances()
                If Not IsNothing(obj("TotalPhysicalMemory")) Then
                    tot = CInt(Convert.ToInt64(obj("TotalPhysicalMemory")) / (1024 * 1024))
                    tot = CInt(IIf((tot Mod 2) <> 0, tot + 1, tot))
                    If num = 0 Or tot > num Then
                        ds.SOS(0).Memory = tot
                        dr.Memory = tot
                    End If
                End If
            Next

            ' record the data only if we've got something valid
            If num <> 0 Then
                ds.Memory.AddMemoryRow(dr)
            End If
        Catch ex As Exception
            dr = ds.Memory.NewMemoryRow
            dr.ID_Memory = ID
            dr.Memory = -1
            dr.Memory_Type = Left("Error: " & ex.Message, ds.Memory.Memory_TypeColumn.MaxLength)
            ds.Memory.AddMemoryRow(dr)
            Misc.ErrorLog(ex, MachineName, "Memory", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about connected Modems
    '
    Public Sub GetModem()
        Dim obj As ManagementObject
        Dim dr As SOSOSDataset.ModemRow

        Try
            wmi.Path.RelativePath = "Win32_POTSModem"
            For Each obj In wmi.GetInstances()
                If obj("DeviceType").ToString <> "Null Modem" Then
                    dr = ds.Modem.NewModemRow
                    dr.ID_Modem = ID
                    dr.Modem_ID = Convert.ToInt32(obj("index"))
                    dr.Modem = Left(obj("Description").ToString, ds.Modem.ModemColumn.MaxLength)
                    dr.Modem_Port = Left(obj("AttachedTo").ToString, ds.Modem.Modem_PortColumn.MaxLength)
                    If obj("DeviceID").ToString.StartsWith("USB") Then
                        dr.Modem_Type = "USB"
                    Else
                        dr.Modem_Type = Left(obj("DeviceType").ToString, ds.Modem.Modem_TypeColumn.MaxLength)
                    End If
                    ds.Modem.AddModemRow(dr)

                    ' yet another item for the SOS summary table
                    ds.SOS(0).Modem = True
                End If
            Next
        Catch ex As Exception
            dr = ds.Modem.NewModemRow
            dr.ID_Modem = ID
            dr.Modem_ID = -1
            dr.Modem = Left("Error: " & ex.Message, ds.Modem.ModemColumn.MaxLength)
            ds.Modem.AddModemRow(dr)
            Misc.ErrorLog(ex, MachineName, "Modem", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Information about the Video monitor.
    ' Note: The monitor's EDID "signature" is stored in the registry
    '
    Public Sub GetMonitor()
        Dim obj As ManagementObject
        Dim PnPDeviceID As String
        Dim cnt As Integer
        Dim dr As SOSOSDataset.MonitorRow

        Try
            cnt = 0
            wmi.Path.RelativePath = "Win32_DesktopMonitor"
            For Each obj In wmi.GetInstances()
                ' not all video cards support the detection of monitors
                dr = ds.Monitor.NewMonitorRow
                dr.ID_Monitor = ID
                dr.Monitor_ID = cnt
                dr.Monitor = Left(obj("MonitorType").ToString, ds.Monitor.MonitorColumn.MaxLength)

                ' try another approach.... Use the EDID data block to get
                ' the monitor's name and to determine the monitor's size
                If Not IsNothing(obj("PnPDeviceID")) Then
                    PnPDeviceID = obj("PnPDeviceID").ToString
                    Misc.Monitor_EDID(reg_hklm, PnPDeviceID, dr)
                End If

                ds.Monitor.AddMonitorRow(dr)
                cnt += 1
            Next

#If TRIM_PHANTOM_MONITORS Then
            ' loop thru the list to see if we have at least one "good" value
            cnt = 0
            For Each dr In ds.Monitor.Rows
                If dr.IsMonitor_SizeNull = False Then
                    cnt += 1
                End If
            Next

            If ds.Monitor.Rows.Count > cnt And cnt > 0 Then
                Dim PendingDeletion As New ArrayList

                ' Let's mark the "bad" rows for deletion
                For Each dr In ds.Monitor
                    If dr.IsMonitor_SizeNull Then
                        PendingDeletion.Add(dr)
                    End If
                Next

                ' Now, let's delete 'em
                For Each dr In PendingDeletion
                    ds.Monitor.RemoveMonitorRow(dr)
                Next
            End If
#End If
        Catch ex As Exception
            dr = ds.Monitor.NewMonitorRow
            dr.ID_Monitor = ID
            dr.Monitor_ID = -1
            dr.Monitor = Left("Error: " & ex.Message, ds.Monitor.MonitorColumn.MaxLength)
            ds.Monitor.AddMonitorRow(dr)
            Misc.ErrorLog(ex, MachineName, "Monitor", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get Manufacturer, Model, and Serial number
    ' Note: Should be called before GetMotherboard()
    '
    Public Sub GetSystemInfo()
        Dim obj As ManagementObject
        Dim dr As SOSOSDataset.SystemInfoRow

        Try
            dr = ds.SystemInfo.NewSystemInfoRow
            dr.ID_System = ID

            wmi.Path.RelativePath = "Win32_ComputerSystem"
            For Each obj In wmi.GetInstances()
                If Not IsNothing(obj("Manufacturer")) Then
                    dr.System_Manufacturer = Left(obj("Manufacturer").ToString, ds.SystemInfo.System_ManufacturerColumn.MaxLength)
                End If
                If Not IsNothing(obj("Model")) Then
                    dr.System_Model = Left(obj("Model").ToString, ds.SystemInfo.System_ModelColumn.MaxLength).Trim
                End If
            Next

            wmi.Path.RelativePath = "Win32_ComputerSystemProduct"
            For Each obj In wmi.GetInstances()
                dr.System_SerialNumber = Left(obj("IdentifyingNumber").ToString, ds.SystemInfo.System_SerialNumberColumn.MaxLength)
                dr.System_UUID = Left(obj("UUID").ToString, ds.SystemInfo.System_UUIDColumn.MaxLength)
            Next

            ' Use the Chassis serial number if this SerialNumber is empty
            ' or blank
            If dr.IsSystem_SerialNumberNull Or (Not dr.IsSystem_SerialNumberNull AndAlso dr.System_SerialNumber = "") Then
                wmi.Path.RelativePath = "Win32_SystemEnclosure"
                For Each obj In wmi.GetInstances()
                    dr.System_SerialNumber = Left(obj("SerialNumber").ToString, ds.SystemInfo.System_SerialNumberColumn.MaxLength)
                Next
            End If

            ' Consider this a "starter value" that will probably be replaced
            ' in the GetMotherboard section below
            If Not dr.IsSystem_ModelNull Then
                ds.SOS(0).PC_Model = dr.System_Model
            End If

            ' Record the data only if we've got a valid Manufacturer field
            If Not dr.IsSystem_ManufacturerNull AndAlso dr.System_Manufacturer.Trim <> "" Then
                ds.SystemInfo.AddSystemInfoRow(dr)
            End If
        Catch ex As Exception
            dr = ds.SystemInfo.NewSystemInfoRow
            dr.ID_System = ID
            dr.System_Manufacturer = Left("Error: " & ex.Message, ds.SystemInfo.System_ManufacturerColumn.MaxLength)
            ds.SystemInfo.AddSystemInfoRow(dr)
            Misc.ErrorLog(ex, MachineName, "SystemInfo", Misc.ErrLogLevel.Errors)
            Exit Sub
        End Try
    End Sub

    '
    ' Motherboard Info
    ' Note: Most servers don't have Win32_BaseBoard values.
    '
    Public Sub GetMotherboard()
        Dim obj As ManagementObject
        Dim temp, Motherboard, Model, names() As String
        Dim dr As SOSOSDataset.MotherboardRow

        Try
            Motherboard = ""
            Model = ""

            wmi.Path.RelativePath = "Win32_BaseBoard"
            For Each obj In wmi.GetInstances()
                dr = ds.Motherboard.NewMotherboardRow
                dr.ID_Motherboard = ID
                dr.Motherboard = Left(obj("Manufacturer").ToString, ds.Motherboard.MotherboardColumn.MaxLength)
                dr.Motherboard_Model = Left(obj("Product").ToString, ds.Motherboard.Motherboard_ModelColumn.MaxLength)
                dr.Motherboard_SerialNumber = Left(obj("SerialNumber").ToString.Trim("."c), ds.Motherboard.Motherboard_SerialNumberColumn.MaxLength)
                If Not IsNothing(obj("Version")) Then
                    dr.Motherboard_Version = Left(obj("Version").ToString, ds.Motherboard.Motherboard_VersionColumn.MaxLength)
                End If
                Motherboard = dr.Motherboard
                Model = dr.Motherboard_Model

                ' Modify the PC_Model field that we started above
                If Not ds.SOS(0).IsPC_ModelNull AndAlso (ds.SOS(0).PC_Model.StartsWith("System ") Or ds.SOS(0).PC_Model = "To Be Filled By O.E.M.") Then
                    ds.SOS(0).PC_Model = Model
                Else
                    ' let's combine the Manufacturer and the model number together
                    ' for the SOS summary table
                    If ds.SystemInfo.Count > 0 Then
                        temp = ds.SystemInfo(0).System_Manufacturer
                        If temp = "System Manufacturer" Then
                            temp = Motherboard
                        End If
                        names = Split(temp, " ")
                        If UBound(names) >= 0 Then
                            If ds.SOS(0).IsPC_ModelNull Then
                                temp = ""
                            Else
                                temp = ds.SOS(0).PC_Model
                            End If
                            If temp.StartsWith("HP ") And temp.Length > 3 Then
                                temp = temp.Substring(3)
                            End If
                            ds.SOS(0).PC_Model = names(0) & " " & temp
                        End If
                    End If
                End If

                ' Save the data only if the Motherboard field is valid. 
                If Not dr.IsMotherboardNull AndAlso dr.Motherboard.Trim <> "" Then
                    ds.Motherboard.AddMotherboardRow(dr)
                End If
            Next
        Catch ex As Exception
            dr = ds.Motherboard.NewMotherboardRow
            dr.ID_Motherboard = ID
            dr.Motherboard = Left("Error: " & ex.Message, ds.Motherboard.MotherboardColumn.MaxLength)
            ds.Motherboard.AddMotherboardRow(dr)
            Misc.ErrorLog(ex, MachineName, "Motherboard", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the mouse
    '
    Public Sub GetMouse()
        Dim obj As ManagementObject
        Dim cnt As Integer
        Dim dr As SOSOSDataset.MouseRow

        Try
            cnt = 0
            wmi.Path.RelativePath = "Win32_PointingDevice"
            For Each obj In wmi.GetInstances()
                dr = ds.Mouse.NewMouseRow
                dr.ID_Mouse = ID
                dr.Mouse_ID = cnt
                dr.Mouse = Left(obj("Description").ToString, ds.Mouse.MouseColumn.MaxLength)
                dr.Mouse_Type = Choose(Convert.ToInt32(obj("PointingType")) + 1, "Other", "Unknown", "Mouse", "Track Ball", "Track Point", "Glide Point", "Touch Pad", "Touch Screen", "Optical Sensor").ToString
                dr.Mouse_Buttons = CInt(obj("NumberOfButtons"))
                If Not IsNothing(obj("DeviceInterface")) Then
                    Select Case Convert.ToInt32(obj("DeviceInterface"))
                        Case 1
                            dr.Mouse_Port = "Other"
                        Case 2
                            dr.Mouse_Port = "Unknown"
                        Case 3
                            dr.Mouse_Port = "Serial"
                        Case 4
                            dr.Mouse_Port = "PS/2"
                        Case 5
                            dr.Mouse_Port = "IR"
                        Case 6
                            dr.Mouse_Port = "HP-HIL"
                        Case 7
                            dr.Mouse_Port = "Bus"
                        Case 8
                            dr.Mouse_Port = "ADB"
                        Case 160
                            dr.Mouse_Port = "Bus DB9"
                        Case 161
                            dr.Mouse_Port = "Bus DIN"
                        Case 162
                            dr.Mouse_Port = "USB"
                        Case Else
                            dr.Mouse_Port = "Unknown"
                    End Select
                End If
#If TRIM_PHANTOM_MICE Then
                If dr.Mouse <> "SMS Virtual Mouse" Then
                    ds.Mouse.AddMouseRow(dr)
                    cnt += 1
                End If
#Else
                ds.Mouse.AddMouseRow(dr)
                cnt += 1
#End If
            Next
        Catch ex As Exception
            dr = ds.Mouse.NewMouseRow
            dr.ID_Mouse = ID
            dr.Mouse_ID = -1
            dr.Mouse = Left("Error: " & ex.Message, ds.Mouse.MouseColumn.MaxLength)
            ds.Mouse.AddMouseRow(dr)
            Misc.ErrorLog(ex, MachineName, "Mouse", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the installed network cards
    '
    Public Sub GetNetAdapter()
        Dim obj As ManagementObject
        Dim nci As Integer
        Dim name As String
        Dim dr As SOSOSDataset.NetAdapterRow

        Try
            nci = 0
            ReDim Netcard(50)

            wmi.Path.RelativePath = "Win32_NetworkAdapter"
            For Each obj In wmi.GetInstances()
                ' we only want *real* LAN adapters... so we skip any WAN
                ' or non-ethernet devices
                If Not IsNothing(obj("Description")) Then
                    name = obj("Description").ToString
                Else
                    name = obj("Name").ToString
                End If
                If Not IsNothing(obj("MACAddress")) AndAlso (Not name.StartsWith("WAN") And Not name.StartsWith("RAS") And name.IndexOf("Miniport") < 0) Then
                    dr = ds.NetAdapter.NewNetAdapterRow
                    dr.ID_NetAdapter = ID
                    dr.NetAdapter_ID = Convert.ToInt32(obj("index"))
                    dr.NetAdapter = Left(obj("Description").ToString, ds.NetAdapter.NetAdapterColumn.MaxLength)
                    dr.NetAdapter_Manufacturer = Left(obj("Manufacturer").ToString, ds.NetAdapter.NetAdapter_ManufacturerColumn.MaxLength)
                    dr.NetAdapter_MAC = Left(obj("MACAddress").ToString, ds.NetAdapter.NetAdapter_MACColumn.MaxLength)
                    ds.NetAdapter.AddNetAdapterRow(dr)

                    ' save the Net Card Index number for later use
                    Netcard(nci) = Convert.ToInt32(obj("index"))
                    nci = nci + 1
                End If
            Next
            ReDim Preserve Netcard(nci - 1)
        Catch ex As Exception
            dr = ds.NetAdapter.NewNetAdapterRow
            dr.ID_NetAdapter = ID
            dr.NetAdapter_ID = -1
            dr.NetAdapter = Left("Error: " & ex.Message, ds.NetAdapter.NetAdapterColumn.MaxLength)
            ds.NetAdapter.AddNetAdapterRow(dr)
            Misc.ErrorLog(ex, MachineName, "NetAdapter", Misc.ErrLogLevel.Errors)
        End Try

        ' If we got nothing above (to include an error message), try this
        ' method instead
        If ds.NetAdapter.Count = 0 Then
            Try
                nci = 0
                ReDim Netcard(50)

                wmi.Path.RelativePath = "Win32_NetworkAdapterConfiguration"
                For Each obj In wmi.GetInstances()
                    If Not IsNothing(obj("IPAddress")) AndAlso UBound(CType(obj("IPAddress"), String())) >= 0 Then
                        dr = ds.NetAdapter.NewNetAdapterRow
                        dr.ID_NetAdapter = ID
                        dr.NetAdapter_ID = Convert.ToInt32(obj("index"))
                        dr.NetAdapter = Left(obj("Description").ToString, ds.NetAdapter.NetAdapterColumn.MaxLength)
                        dr.NetAdapter_MAC = Left(obj("MACAddress").ToString, ds.NetAdapter.NetAdapter_MACColumn.MaxLength)
                        ds.NetAdapter.AddNetAdapterRow(dr)

                        ' save the Net Card Index number for later use
                        Netcard(nci) = Convert.ToInt32(obj("index"))
                        nci = nci + 1
                    End If
                Next
                ReDim Preserve Netcard(nci - 1)
            Catch ex As Exception
                dr = ds.NetAdapter.NewNetAdapterRow
                dr.ID_NetAdapter = ID
                dr.NetAdapter_ID = -1
                dr.NetAdapter = Left("Error: " & ex.Message, ds.NetAdapter.NetAdapterColumn.MaxLength)
                ds.NetAdapter.AddNetAdapterRow(dr)
                Misc.ErrorLog(ex, MachineName, "NetAdapter", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the network configuration associated with each of the network
    ' cards identified above
    '
    Public Sub GetNetConfig()
        ' We only need to do this if we have at least one Net Card Index
        ' value (created above).
        If Netcard.Length >= 0 Then
            Dim obj As ManagementObject
            Dim num, cnt As Integer
            Dim dr As SOSOSDataset.NetConfigRow

            Try
                wmi.Path.RelativePath = "Win32_NetworkAdapterConfiguration"
                For Each obj In wmi.GetInstances()
                    For cnt = 0 To UBound(Netcard)
                        ' is it a match for the index in the NetAdapter section above?
                        If Convert.ToInt32(obj("Index")) = Netcard(cnt) Then
                            dr = ds.NetConfig.NewNetConfigRow
                            dr.ID_NetConfig = ID
                            dr.NetConfig_ID = Netcard(cnt)
                            ' "Caption" is the best name, but it's typically
                            ' enclosed in square brackets
                            num = InStr(obj("Caption").ToString, "] ")
                            If num > 0 Then
                                num = CInt(IIf(num > ds.NetConfig.NetConfigColumn.MaxLength - 2, ds.NetConfig.NetConfigColumn.MaxLength - 2, num))
                                dr.NetConfig = Mid(obj("Caption").ToString, num + 2)
                            Else
                                dr.NetConfig = Left(obj("Description").ToString, ds.NetConfig.NetConfigColumn.MaxLength)
                            End If
                            dr.NetConfig_DHCP = False
                            If Not IsNothing(obj("DHCPEnabled")) Then
                                dr.NetConfig_DHCP = CBool(obj("DHCPEnabled"))
                            End If
                            If Not IsNothing(obj("IPAddress")) AndAlso UBound(CType(obj("IPAddress"), String())) >= 0 Then
                                dr.NetConfig_IP = Misc.Format_IP(CType(obj("IPAddress"), String())(0))
                            End If
                            If Not IsNothing(obj("IPSubnet")) AndAlso UBound(CType(obj("IPSubnet"), String())) >= 0 Then
                                dr.NetConfig_Subnet = Misc.Format_IP(CType(obj("IPSubnet"), String())(0))
                            End If
                            If Not IsNothing(obj("DefaultIPGateway")) AndAlso UBound(CType(obj("DefaultIPGateway"), String())) >= 0 Then
                                dr.NetConfig_Router = Misc.Format_IP(CType(obj("DefaultIPGateway"), String())(0))
                            End If
                            If Not IsNothing(obj("DNSHostName")) Then
                                dr.NetConfig_DNSName = Left(obj("DNSHostName").ToString, ds.NetConfig.NetConfig_DNSNameColumn.MaxLength).ToLower
                            End If
                            If Not IsNothing(obj("DNSDomain")) Then
                                dr.NetConfig_DNSDomain = Left(obj("DNSDomain").ToString, ds.NetConfig.NetConfig_DNSDomainColumn.MaxLength)
                            End If
                            If Not IsNothing(obj("DNSServerSearchOrder")) AndAlso UBound(CType(obj("DNSServerSearchOrder"), String())) >= 0 Then
                                dr.NetConfig_DNS1 = Misc.Format_IP(CType(obj("DNSServerSearchOrder"), String())(0))
                            End If
                            If Not IsNothing(obj("DNSServerSearchOrder")) AndAlso UBound(CType(obj("DNSServerSearchOrder"), String())) >= 1 Then
                                dr.NetConfig_DNS2 = Misc.Format_IP(CType(obj("DNSServerSearchOrder"), String())(1))
                            End If
                            If Not IsNothing(obj("WINSPrimaryServer")) Then
                                dr.NetConfig_WINS1 = Misc.Format_IP(obj("WINSPrimaryServer").ToString)
                            End If
                            If Not IsNothing(obj("WINSSecondaryServer")) Then
                                dr.NetConfig_WINS2 = Misc.Format_IP(obj("WINSSecondaryServer").ToString)
                            End If
                            ds.NetConfig.AddNetConfigRow(dr)
                        End If
                    Next
                Next

                ' If there is more than one network card, try to figure out
                ' which one we should use in the SOS summary table.
                For cnt = 0 To ds.NetConfig.Count - 1
                    If Not ds.NetConfig(cnt).IsNetConfig_IPNull AndAlso (ds.NetConfig(cnt).NetConfig_IP <> "000.000.000.000" And ds.NetConfig(cnt).NetConfig_IP.Length > 0) Then
                        ' shorten it up a wee bit.  Some descriptions have some 
                        ' text in quotes... if so, whack it (if it occurs at least
                        ' half way thru the string)
                        num = InStr(ds.NetAdapter(cnt).NetAdapter, " (")
                        If num > ds.NetAdapter(cnt).NetAdapter.Length / 2 Then
                            ds.SOS(0).NetCard = Left(ds.NetAdapter(cnt).NetAdapter, num - 1)
                        Else
                            ds.SOS(0).NetCard = ds.NetAdapter(cnt).NetAdapter
                        End If
                        ds.SOS(0).IP_Address = ds.NetConfig(cnt).NetConfig_IP
                        ds.SOS(0).DHCP = ds.NetConfig(cnt).NetConfig_DHCP
                        ' The MAC address of the "current" adapter may not be the same
                        ' as the MAC used in the ID field.  This is OK, since we don't
                        ' want duplicate entries in the database based upon which card
                        ' is current.
                        ds.SOS(0).MAC_Address = ds.NetAdapter(cnt).NetAdapter_MAC
                        Exit For
                    End If
                Next

                ' one last chance for the Summary table...
                If ds.SOS(0).IsNetCardNull Then
                    If Not ds.NetAdapter(0).IsNetAdapterNull Then
                        ds.SOS(0).NetCard = ds.NetAdapter(0).NetAdapter
                    End If
                    If Not ds.NetAdapter(0).IsNetAdapter_MACNull Then
                        ds.SOS(0).MAC_Address = ds.NetAdapter(0).NetAdapter_MAC
                    End If
                End If
            Catch ex As Exception
                dr = ds.NetConfig.NewNetConfigRow
                dr.ID_NetConfig = ID
                dr.NetConfig_ID = -1
                dr.NetConfig_DHCP = False
                dr.NetConfig = Left("Error: " & ex.Message, ds.NetConfig.NetConfigColumn.MaxLength)
                ds.NetConfig.AddNetConfigRow(dr)
                Misc.ErrorLog(ex, MachineName, "NetConfig", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get detailed information about the Operating System
    '
    Public Sub GetOS()
        Dim obj As ManagementObject
        Dim cnt As Integer
        Dim temp, csd As String
        Dim dr As SOSOSDataset.OSRow

        Try
            wmi.Path.RelativePath = "Win32_OperatingSystem"
            For Each obj In wmi.GetInstances()
                dr = ds.OS.NewOSRow
                dr.ID_OS = ID

                ' "Name" is the best OS Name... but it has a bizzare format so
                ' we have to parse it
                cnt = InStr(obj("Name").ToString, "|")
                If cnt > 0 Then
                    temp = Left(obj("Name").ToString, cnt - 1)
                Else
                    temp = obj("Caption").ToString
                End If
                csd = ""
                If Not IsNothing(obj("csdversion")) Then
                    csd = obj("csdversion").ToString
                End If
                dr.OS = Left(Trim(temp & " " & csd), ds.OS.OSColumn.MaxLength)
                dr.OS_Version = Left(obj("Version").ToString, ds.OS.OS_VersionColumn.MaxLength)
                If Not IsNothing(obj("installdate")) Then
                    dr.OS_InstallDate = ManagementDateTimeConverter.ToDateTime(obj("installdate").ToString)
                End If
                dr.OS_RegUser = Left(obj("registereduser").ToString, ds.OS.OS_RegUserColumn.MaxLength)
                If Not IsNothing(obj("organization")) Then
                    dr.OS_Organization = Left(obj("organization").ToString, ds.OS.OS_OrganizationColumn.MaxLength)
                End If
                dr.OS_SerialNumber = Left(obj("SerialNumber").ToString, ds.OS.OS_SerialNumberColumn.MaxLength)

                ' shorten it up a bit for the SOS summary table
                If temp.StartsWith("Microsoft Windows NT") Then
                    ds.SOS(0).OS = Left(Replace(temp, "Microsoft ", "") & " " & Format(OSVer, "0.0") & " " & Trim(Replace(obj("csdversion").ToString, "Service Pack", "SP")), ds.OS.OSColumn.MaxLength)
                Else
                    ds.SOS(0).OS = Left(Replace(temp, "Microsoft ", "") & " " & Trim(Replace(csd, "Service Pack", "SP")), ds.OS.OSColumn.MaxLength)
                End If

                ' another item for the SOS summary table
                ds.SOS(0).OS_Build = dr.OS_Version

                ds.OS.AddOSRow(dr)
            Next
        Catch ex As Exception
            dr = ds.OS.NewOSRow
            dr.ID_OS = ID
            dr.OS = Left("Error: " & ex.Message, ds.OS.OSColumn.MaxLength)
            ds.OS.AddOSRow(dr)
            Misc.ErrorLog(ex, MachineName, "OS", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' This is a collection of a lot of things that just didn't fit 
    ' anywhere else
    '
    Public Sub GetPC()
        Dim cnt As Integer
        Dim obj As ManagementObject
        Dim temp, ResourceDomain, LastLoginFile As String
        Dim key As RegistryKey
        Dim isAD As Boolean
        Dim dr As SOSOSDataset.PCRow

        Try
            dr = ds.PC.NewPCRow
            dr.ID_PC = ID

            ResourceDomain = ""
            wmi.Path.RelativePath = "Win32_ComputerSystem"
            For Each obj In wmi.GetInstances()
                dr.PC = Left(obj("Name").ToString, ds.PC.PCColumn.MaxLength)
                dr.PC_Timezone = CInt(obj("CurrentTimeZone"))
                ' for Win9x clients
                If IsNothing(obj("DomainRole")) Then
                    key = reg_hklm.OpenSubKey("System\CurrentControlSet\Services\VxD\VNETSUP", False)
                    If Not IsNothing(key) AndAlso Not IsNothing(key.GetValue("Workgroup")) Then
                        ResourceDomain = key.GetValue("Workgroup").ToString.ToUpper
                        key.Close()
                    End If
                Else
                    ResourceDomain = obj("Domain").ToString
                    ' some systems record the NetBIOS version of the
                    ' name, others the full DNS version
                    If ResourceDomain.IndexOf(".") < 0 And DomainRole <> 0 And DomainRole <> 2 Then
                        key = reg_hklm.OpenSubKey("SYSTEM\CurrentControlSet\Services\Tcpip\Parameters", False)
                        If Not IsNothing(key) AndAlso Not IsNothing(key.GetValue("Domain")) Then
                            temp = key.GetValue("Domain").ToString
                            If temp <> "" Then
                                ResourceDomain = temp
                            End If
                            key.Close()
                        End If
                    End If
                End If
                dr.PC_DomainName = Left(ResourceDomain, ds.PC.PC_DomainNameColumn.MaxLength)
                dr.PC_DomainRole = Choose(DomainRole + 1, "WorkGroup", "Domain Member", "WorkGroup Server", "Domain Server", "BDC", "PDC").ToString
                dr.PC_NetServices = CBool(obj("NetworkServerModeEnabled"))
                dr.PC_SystemType = Left(obj("SystemType").ToString, ds.PC.PC_SystemTypeColumn.MaxLength)
            Next

            ' Get Daylight savings time
            dr.PC_DSTEnabled = False
            key = reg_hklm.OpenSubKey("SYSTEM\CurrentControlSet\Control\TimeZoneInformation", False)
            If Not IsNothing(key) Then
                Dim regbuf As Object
                regbuf = key.GetValue("DaylightBias", 0)
                ' for Win9x, it's a byte array
                If regbuf.GetType Is GetType(Byte()) Then
                    dr.PC_DSTEnabled = CBool(CType(regbuf, Byte())(0))
                Else
                    dr.PC_DSTEnabled = CBool(regbuf)
                End If
                key.Close()
            End If

            ' Let's use the "downlevel" (NetBIOS) name for the ResourceDomain
            ' (PC_DomainName is still the FQDN)
            isAD = False
            If DomainRole <> 0 And DomainRole <> 2 And ResourceDomain.IndexOf(".") > 0 Then
                Dim part As DirectoryEntry
                Dim parts As New DirectoryEntry
                Dim names(), NetBIOSDomain As String

                NetBIOSDomain = ""
                Try
                    ' build a CN style domain name
                    names = ResourceDomain.Split("."c)
                    temp = "DC=" & Join(names, ",DC=")

                    ' search the AD containers for one that matches the domain's DN.
                    parts.Path = "LDAP://CN=Partitions,CN=Configuration," & temp
                    For Each part In parts.Children
                        If part.Properties("nCName").Value.ToString.ToLower = temp.ToLower Then
                            NetBIOSDomain = part.Properties("nETBIOSName").Value.ToString
                            isAD = True
                            Exit For
                        End If
                    Next
                Catch ldap_ex As System.Runtime.InteropServices.COMException
                    ' There may be some cases where permissions on the AD Partitions
                    ' container are so restrictive that we can't get the NetBIOS domain
                    ' names.  So, we fake it
                    NetBIOSDomain = Left(ResourceDomain.Split("."c)(0).ToUpper, 14)
                    isAD = True
                End Try
                If NetBIOSDomain <> "" Then
                    dr.PC_DomainNameNB = NetBIOSDomain
                End If
            End If

            ' get boot time
            wmi.Path.RelativePath = "Win32_OperatingSystem"
            For Each obj In wmi.GetInstances()
                If Not IsNothing(obj("LastBootUpTime")) Then
                    dr.PC_LastBoot = ManagementDateTimeConverter.ToDateTime(obj("LastBootUpTime").ToString)
                End If
            Next

            ' Get the last login time (by anybody on this PC)
            ' Note: We have a login batch that creates this file...
            temp = My.Settings.LastLoginFile
            If temp <> "" Then
                ' in case you didn't read the documentation :)
                If temp.ToUpper.StartsWith("\") Then
                    temp = SystemDrive & temp
                End If
                LastLoginFile = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
                If File.Exists(Misc.ConvUNC(RemotePC, LastLoginFile)) Then
                    dr.PC_LastLogin = File.GetLastWriteTime(Misc.ConvUNC(RemotePC, LastLoginFile))
                End If
            End If

            ' count the number of desktops.  A crude way to determining
            ' if this is a "gang" PC or not.
            ' Note: this get replaced in the GetProfiles() method
            cnt = 0
            wmi.Path.RelativePath = "Win32_Desktop"
            For Each obj In wmi.GetInstances()
                cnt += 1
            Next
            dr.PC_NumUsers = cnt

            ' get the chassis type(s)
            Dim buf, bufs() As UInt16
            temp = ""
            wmi.Path.RelativePath = "Win32_SystemEnclosure"
            For Each obj In wmi.GetInstances()
                bufs = CType(obj("ChassisTypes"), UInt16())
                For Each buf In bufs
                    If temp <> "" Then
                        temp = temp & ", "
                    End If
                    temp = temp & Choose(Convert.ToInt32(buf), "Other", "Unknown", "Desktop", "Low Profile", "Pizza Box", "Mini Tower", "Tower", "Portable", "Laptop", "Notebook", "Hand Held", "Docking Station", "All in One", "Sub Notebook", "Space Saving", "Lunch Box", "Main System", "Expansion", "SubChassis", "Bus Expansion", "Peripheral", "Storage", "Rank Mount", "Sealed").ToString
                Next
            Next
            dr.PC_ChassisType = Left(temp, ds.PC.PC_ChassisTypeColumn.MaxLength)

            ' Active Directory "full path" (the Distinguished Name)
            If isAD Then
                key = reg_hklm.OpenSubKey("SOFTWARE\Microsoft\Windows\CurrentVersion\Group Policy\State\Machine", False)
                If Not IsNothing(key) AndAlso Not IsNothing(key.GetValue("Distinguished-Name")) Then
                    dr.PC_ADPath = Left(key.GetValue("Distinguished-Name").ToString, ds.PC.PC_ADPathColumn.MaxLength)
                    key.Close()
                End If
                If dr.IsPC_ADPathNull Then
                    ' OK... we have to do it the hard way.  We do an LDAP
                    ' search for this PC's distinguished name
                    Dim root As DirectoryEntry
                    Dim ad As DirectorySearcher
                    Dim res As SearchResult

                    Try
                        root = New DirectoryEntry("LDAP://" & ResourceDomain)
                        ad = New DirectorySearcher(root, "(&(objectclass=computer)(cn=" & MachineName & "))")
                        res = ad.FindOne()
                        If Not IsNothing(res) Then
                            dr.PC_ADPath = Left(res.Properties("distinguishedName")(0).ToString, ds.PC.PC_ADPathColumn.MaxLength)
                        End If
                    Catch ldap_ex As System.Runtime.InteropServices.COMException
                        ' ignore
                    End Try
                End If
            End If

            ' The SID value came from the enumerating of the local accounts.
            If SID <> "" Then
                dr.PC_SID = SID
            Else
                ' The SID for domain controllers is a bit different, we must
                ' use yet another LDAP query
                If DomainRole > 3 And Not dr.IsPC_ADPathNull Then
                    Dim root As DirectoryEntry
                    Dim RawSID() As Byte
                    Dim names() As String

                    Try
                        root = New DirectoryEntry("LDAP://" & dr.PC_ADPath)
                        RawSID = CType(root.Properties("objectSid").Value, Byte())

                        ' we want all but the last part
                        names = Misc.SIDtoString(RawSID).Split("-"c)
                        ReDim Preserve names(UBound(names) - 1)

                        dr.PC_SID = Join(names, "-")
                    Catch ldap_ex As System.Runtime.InteropServices.COMException
                        ' ignore
                    End Try
                End If
            End If

            ' Can't get the Comment with WMI... so, we have to poke around
            ' the registry
            key = reg_hklm.OpenSubKey("SYSTEM\CurrentControlSet\Services\lanmanserver\parameters", False)
            If IsNothing(key) Then
                ' Win9x
                key = reg_hklm.OpenSubKey("System\CurrentControlSet\Services\VxD\VNETSUP", False)
                If Not IsNothing(key) AndAlso Not IsNothing(key.GetValue("comment")) Then
                    dr.PC_Comment = Left(key.GetValue("comment").ToString, ds.PC.PC_CommentColumn.MaxLength)
                    key.Close()
                End If
            Else
                If Not IsNothing(key.GetValue("srvcomment")) Then
                    dr.PC_Comment = Left(key.GetValue("srvcomment").ToString, ds.PC.PC_CommentColumn.MaxLength)
                End If
                key.Close()
            End If

            ' knowing the HAL is useful when cloning PCs
            key = reg_hklm.OpenSubKey("HARDWARE\RESOURCEMAP\Hardware Abstraction Layer", False)
            If Not IsNothing(key) Then
                dr.PC_HAL = Left(key.GetSubKeyNames(0), ds.PC.PC_HALColumn.MaxLength)
                key.Close()
            End If

            ' Get the date the PC was first joined to the domain 
            temp = Misc.ConvUNC(RemotePC, SystemRoot & "\system32\config\netlogon.ftl")
            If File.Exists(temp) Then
                dr.PC_JoinedDomain = File.GetCreationTime(temp)
            End If
            ' for domain controllers, it a bit different
            key = reg_hklm.OpenSubKey("System\CurrentControlSet\Services\Netlogon\Parameters", False)
            If Not IsNothing(key) Then
                temp = key.GetValue("SYSVOL", "").ToString
                If temp <> "" Then
                    temp = Misc.ConvUNC(RemotePC, temp)
                    If Directory.Exists(temp) Then
                        dr.PC_JoinedDomain = Directory.GetCreationTime(temp)
                    End If
                End If
                key.Close()
            End If

            ' Update a bunch of values in the SOS Summary table
            ds.SOS(0).PC_Domain = dr.PC_DomainName
            If Not dr.IsPC_CommentNull Then
                ds.SOS(0).PC_Comment = dr.PC_Comment
            End If
            ds.SOS(0).Net_Services = dr.PC_NetServices

            ds.PC.AddPCRow(dr)
        Catch ex As Exception
            dr = ds.PC.NewPCRow
            dr.ID_PC = ID
            dr.PC_DSTEnabled = False
            dr.PC_NetServices = False
            dr.PC = Left("Error: " & ex.Message, ds.PC.PCColumn.MaxLength)
            ds.PC.AddPCRow(dr)
            Misc.ErrorLog(ex, MachineName, "PC", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' The Permissions table is only populated for items contained in the
    ' Shares table
    ' Note: Only Administrators can get Permissions
    '
    Public Sub GetPermissions()
        If ds.Shares.Count > 0 And IsWin32 Then
            Dim opt As InvokeMethodOptions
            Dim out, sd, ace, trustee, dacl() As ManagementBaseObject
            Dim perm_cnt As Integer
            Dim ShareTarget As String
            Dim defaulted As Boolean
            Dim dr_sh As SOSOSDataset.SharesRow
            Dim dr As SOSOSDataset.PermissionsRow

            ShareTarget = ""
            Try
                For Each dr_sh In ds.Shares
                    ShareTarget = dr_sh.Share_Desc
                    ' We can skip rows with error messages and rows without a Share_Path.
                    ' A "null" Share_Path is a good clue that the user is not an
                    ' administrator, so there's no need to try to get the permissions.
                    If Not ShareTarget.StartsWith("Error:") And Not dr_sh.IsShare_PathNull Then
                        defaulted = False
                        out = Nothing

                        ' As a hold-over from WinNT, a share that has no Security Descriptor
                        ' is the same as the default permissions... so we fake it
                        Try
                            wmi.Path.RelativePath = "Win32_LogicalShareSecuritySetting.Name=" & Chr(34) & ShareTarget & Chr(34)
                            opt = New InvokeMethodOptions(Nothing, New TimeSpan(0, 0, 0, 5))
                            out = wmi.InvokeMethod("GetSecurityDescriptor", Nothing, opt)
                        Catch mex As System.Management.ManagementException
                            ' Is this the expected "Not Found" error?
                            If mex.ErrorCode = ManagementStatus.NotFound Then
                                dr = ds.Permissions.NewPermissionsRow()
                                dr.ID_Perm = ID
                                dr.Perm_ID = perm_cnt
                                dr.Perm_Target = ShareTarget
                                dr.Perm_Type = "Allow"
                                dr.Perm_Name = "Everyone"
                                dr.Perm_Permission = "Full Control"
                                ds.Permissions.AddPermissionsRow(dr)
                                defaulted = True
                                perm_cnt += 1
                            Else
                                Throw
                            End If
                        End Try

                        If Not defaulted Then
                            sd = CType(out("Descriptor"), ManagementBaseObject)
                            dacl = CType(sd("DACL"), ManagementBaseObject())
                            For Each ace In dacl
                                dr = ds.Permissions.NewPermissionsRow()
                                dr.ID_Perm = ID
                                dr.Perm_ID = perm_cnt
                                dr.Perm_Target = ShareTarget
                                dr.Perm_Type = Choose((Convert.ToInt64(ace("AceType")) + 1), "Allow", "Deny", "Audit").ToString
                                trustee = CType(ace("Trustee"), ManagementBaseObject)
                                If Not IsNothing(trustee("Domain")) Then
                                    dr.Perm_Name = Left(trustee("Domain").ToString & "\" & trustee("Name").ToString, ds.Permissions.Perm_NameColumn.MaxLength)
                                Else
                                    If Not IsNothing(trustee("Name")) Then
                                        dr.Perm_Name = Left(trustee("Name").ToString, ds.Permissions.Perm_NameColumn.MaxLength)
                                    End If
                                End If
                                dr.Perm_Permission = Left(Misc.MaskToString(Convert.ToInt32(ace("AccessMask"))), ds.Permissions.Perm_PermissionColumn.MaxLength)
                                ds.Permissions.AddPermissionsRow(dr)
                                perm_cnt += 1
                            Next
                        End If
                    End If
                Next
            Catch ex As Exception
                dr = ds.Permissions.NewPermissionsRow()
                dr.ID_Perm = ID
                dr.Perm_ID = -1
                ' it's possible that an early error wouldn't populate the
                ' ShareTarget field (which is a part of the foreign key)
                If ShareTarget = "" Then
                    ShareTarget = ds.Shares(0).Share_Desc
                End If
                dr.Perm_Target = ShareTarget
                dr.Perm_Type = Left("Error: " & ex.Message, ds.Permissions.Perm_TypeColumn.MaxLength)
                ds.Permissions.AddPermissionsRow(dr)
                perm_cnt += 1
                Misc.ErrorLog(ex, MachineName, "Permissions", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Ports (actually, just the number of each type of port)
    '
    Public Sub GetPorts()
        Dim obj As ManagementObject
        Dim cnt As Integer
        Dim dr As SOSOSDataset.PortsRow

        Try
            dr = ds.Ports.NewPortsRow()
            dr.ID_Ports = ID

            cnt = 0
            wmi.Path.RelativePath = "Win32_SerialPort"
            For Each obj In wmi.GetInstances()
                cnt += 1
            Next
            dr.Ports_Serial = cnt

            cnt = 0
            wmi.Path.RelativePath = "Win32_ParallelPort"
            For Each obj In wmi.GetInstances()
                cnt += 1
            Next
            dr.Ports_Parallel = cnt

            cnt = 0
            wmi.Path.RelativePath = "Win32_InfraredDevice"
            For Each obj In wmi.GetInstances()
                cnt += 1
            Next
            dr.Ports_IR = cnt

            cnt = 0
            wmi.Path.RelativePath = "Win32_PCMCIAController"
            For Each obj In wmi.GetInstances()
                cnt += 1
            Next
            dr.Ports_PCMCIA = cnt
            If cnt > 0 Then
                ds.SOS(0).PCMCIA = True
            End If

            cnt = 0
            wmi.Path.RelativePath = "Win32_USBController"
            For Each obj In wmi.GetInstances()
                cnt += 1
            Next
            dr.Ports_USB = cnt
            If cnt > 0 Then
                ds.SOS(0).USB = True
            End If

            cnt = 0
            wmi.Path.RelativePath = "Win32_1394Controller"
            For Each obj In wmi.GetInstances()
                cnt += 1
            Next
            dr.Ports_Firewire = cnt

            ds.Ports.AddPortsRow(dr)
        Catch ex As Exception
            dr = ds.Ports.NewPortsRow()
            dr.ID_Ports = ID
            dr.Ports_Message = Left("Error: " & ex.Message, ds.Ports.Ports_MessageColumn.MaxLength)
            ds.Ports.AddPortsRow(dr)
            Misc.ErrorLog(ex, MachineName, "Ports", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about installed printers (includes networked printers).
    '
    Public Sub GetPrinters()
        Dim obj As ManagementObject
        Dim cnt As Integer
        Dim temp As String
        Dim dr As SOSOSDataset.PrintersRow

        Try
            cnt = 0
            wmi.Path.RelativePath = "Win32_Printer"
            For Each obj In wmi.GetInstances()
                dr = ds.Printers.NewPrintersRow
                dr.ID_Printer = ID
                dr.Printer_ID = cnt
                dr.Printer = Left(obj("Caption").ToString, ds.Printers.PrinterColumn.MaxLength)
                dr.Printer_Port = Left(obj("PortName").ToString, ds.Printers.Printer_PortColumn.MaxLength)
                ' Is this a *real* local printer?
                temp = Left(dr.Printer_Port, 3)
                If (temp = "LPT" Or temp = "COM" Or temp = "USB" Or temp = "DOT") And dr.Printer.StartsWith("Acrobat") = False And dr.Printer.StartsWith("Adobe") = False And dr.Printer.StartsWith("Generic") = False And dr.Printer.StartsWith("\\") = False Then
                    If Not IsNothing(obj("ShareName")) Then
                        dr.Printer_SharedAs = Left(obj("ShareName").ToString, ds.Printers.Printer_SharedAsColumn.MaxLength)
                    End If
                    ds.SOS(0).Printer = dr.Printer
                End If
                ds.Printers.AddPrintersRow(dr)
                cnt += 1
            Next
        Catch ex As Exception
            dr = ds.Printers.NewPrintersRow
            dr.ID_Printer = ID
            dr.Printer_ID = -1
            dr.Printer = Left("Error: " & ex.Message, ds.Printers.PrinterColumn.MaxLength)
            ds.Printers.AddPrintersRow(dr)

            ' Check to see if this is an error that is fairly common.  If so,
            ' let's not clog up the error log with it.
            If ex.GetType Is GetType(System.Management.ManagementException) Then
                If CType(ex, ManagementException).ErrorCode = ManagementStatus.Failed Then
                    Misc.ErrorLog(ex, MachineName, "Printers", Misc.ErrLogLevel.Full)
                    Exit Try
                End If
            End If
            Misc.ErrorLog(ex, MachineName, "Printers", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the currently running processes
    ' Note: This is about as close to "big brother" as we get.  On the other
    ' hand, if this runs during a login batch file, then the user won't have
    ' had time to launch any applications.
    '
    Public Sub GetProcesses()
        ' Not available on Win9x
        If IsWin32 Then
            Dim obj As ManagementObject
            Dim names(0) As String
            Dim ret As Integer
            Dim dr As SOSOSDataset.ProcessesRow

            Try
                wmi.Path.RelativePath = "Win32_Process"
                For Each obj In wmi.GetInstances()
                    dr = ds.Processes.NewProcessesRow
                    dr.ID_Process = ID
                    dr.ProcessID = Convert.ToInt32(obj("ProcessID"))
                    dr.Process = Left(obj("Description").ToString, ds.Processes.ProcessColumn.MaxLength)
                    If Not IsNothing(obj("ExecutablePath")) Then
                        dr.Process_Path = Left(obj("ExecutablePath").ToString, ds.Processes.Process_PathColumn.MaxLength)
                    End If
                    ' get the owner of the process
                    Try
                        ret = Convert.ToInt32(obj.InvokeMethod("GetOwner", names))
                        If ret = 0 Then
                            dr.Process_Owner = Left(names(0), ds.Processes.Process_OwnerColumn.MaxLength)
                        End If
                    Catch mex As System.Management.ManagementException When mex.ErrorCode = ManagementStatus.NotFound
                        ' This is a common error where the process has exited
                        ' and the GetOwner method can't find the process.
                    End Try
                    dr.Process_Memory = CInt(Convert.ToInt64(obj("WorkingSetSize")) / 1024)
                    ds.Processes.AddProcessesRow(dr)
                Next

                Misc.DataTableSort(ds.Processes, "ProcessID")
            Catch ex As Exception
                dr = ds.Processes.NewProcessesRow
                dr.ID_Process = ID
                dr.ProcessID = -1
                dr.Process = Left("Error: " & ex.Message, ds.Processes.ProcessColumn.MaxLength)
                ds.Processes.AddProcessesRow(dr)
                Misc.ErrorLog(ex, MachineName, "Processes", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get information about the User Profiles and mount all of the User's
    ' registry hives.
    ' Note: The registry hives mounted here are used in: GetEmail, GetInternet,
    ' GetMapped, GetShares, and GetStartup.  We do not update the "reference
    ' count" for the hive, so it's possible that some other program might
    ' umount the hive before we read from it.
    '
    Public Sub GetProfiles()
        If IsWin32 Then
            Dim DefaultUserProfile, MountPoint, temp, ImagePath, ProfileName, UserID, LoadedHives() As String
            Dim key, subkey, testkey As RegistryKey
            Dim flags, num_users As Integer
            Dim ld As LongDate
            Dim RawSID() As Byte
            Dim admin As Boolean
            Dim dr As SOSOSDataset.ProfilesRow

            ImagePath = ""
            MountPoint = ""
            Try
                ' the mounting of registry hives is a tightly held privilege
                ' that is not enabled by default (even for administrators)
                admin = Misc.IsAdmin()

                ' make a list of hives that are currently loaded
                LoadedHives = reg_hku.GetSubKeyNames

                ' get the listing of profiles from the registry
                key = reg_hklm.OpenSubKey("SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProfileList", False)
                If Not IsNothing(key) Then
                    ' loop thru each subkey
                    num_users = 0
                    For Each MountPoint In key.GetSubKeyNames
                        subkey = key.OpenSubKey(MountPoint, False)
                        flags = CInt(subkey.GetValue("Flags", -1))
                        RawSID = CType(subkey.GetValue("SID", Nothing), Byte())

                        ' for WinVista, the flags trick doesn't work to
                        ' weed out the non-numan keys
                        If OSVer >= 6.0 And IsNothing(RawSID) Then
                            flags = -1
                        End If

                        ' we only want "human" profiles (WinNT doesn't have
                        ' this flag, but it doesn't list non-human keys)
                        If flags = 0 Or OSVer = 4.0 Then
                            num_users += 1
                            temp = subkey.GetValue("ProfileImagePath", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString
                            ImagePath = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
                            If Directory.Exists(Misc.ConvUNC(RemotePC, ImagePath)) Then
                                ld.date_hi = CInt(subkey.GetValue("ProfileLoadTimeHigh", 0))
                                ld.date_lo = CInt(subkey.GetValue("ProfileLoadTimeLow", 0))
                                ProfileName = Path.GetFileName(ImagePath)

                                ' It's possible that you could have a 
                                ' duplicate profile, since the path may have
                                ' been used by the same user with an older
                                ' SID.  In that case zap the older record
                                Dim dup_dr As SOSOSDataSet.ProfilesRow = ds.Profiles.FindByID_ProfileProfile(ID, Left(ProfileName, ds.Profiles.ProfileColumn.MaxLength))
                                If Not IsNothing(dup_dr) Then
                                    ds.Profiles.RemoveProfilesRow(dup_dr)
                                End If

                                If Not IsNothing(RawSID) Then
                                    UserID = Misc.SidToName(RemotePC, RawSID)
                                Else
                                    ' If we don't have an imbedded SID, then
                                    ' we use the key name instead
                                    UserID = Misc.SidToName(RemotePC, MountPoint)
                                End If

                                ' Let's mount the user's hive (if it isn't
                                ' already mounted)
                                If admin AndAlso Array.IndexOf(LoadedHives, MountPoint) < 0 Then
                                    ' Since we're doing this remotely, the path
                                    ' is relative to the remote PC
                                    Try
                                        Misc.MountHive(RemotePC, MountPoint, ImagePath & "\NTUSER.DAT")
                                        ' we tag the MountPoint with an "*" if
                                        ' we mounted it (used in the Dispose to
                                        ' unmount the hives)
                                        ProfileList.Add(ProfileName, "*" & MountPoint)
                                    Catch ex As Exception
                                        ' I don't care why
                                    End Try
                                End If

                                ' add this ProfileName to the list
                                If Not ProfileList.ContainsKey(ProfileName) Then
                                    ProfileList.Add(ProfileName, MountPoint)
                                End If

                                dr = ds.Profiles.NewProfilesRow
                                dr.ID_Profile = ID
                                dr.Profile = Left(ProfileName, ds.Profiles.ProfileColumn.MaxLength)
                                dr.Profile_Path = Left(ImagePath, ds.Profiles.Profile_PathColumn.MaxLength)
                                If UserID <> "" Then
                                    dr.Profile_Owner = Left(UserID, ds.Profiles.Profile_OwnerColumn.MaxLength)
                                End If

                                If OSVer = 4.0 Or admin = False Then
                                    ' This is a bit crude, but it's better
                                    ' than nothing
                                    dr.Profile_IsAdmin = False
                                    If Not IsNothing(ds.Admins.FindByID_AdminAdmin_UserID(ID, UserID)) Then
                                        dr.Profile_IsAdmin = True
                                    End If
                                Else
                                    dr.Profile_IsAdmin = Misc.IsRemoteAdmin(MachineName, reg_hku, UserID)
                                End If
                                If ld.Long64 > 0 Then
                                    dr.Profile_LastLoad = Date.FromFileTime(ld.Long64)
                                Else
                                    ' ProfileLoadTime is not supported in WinNT
                                    ' or WinVista, so we use the mod date of
                                    ' the NTUSER.DAT file instead.
                                    If File.Exists(Misc.ConvUNC(RemotePC, ImagePath & "\NTUSER.DAT")) Then
                                        dr.Profile_LastLoad = File.GetLastWriteTime(Misc.ConvUNC(RemotePC, ImagePath & "\NTUSER.DAT"))
                                    End If
                                End If

                                ds.Profiles.AddProfilesRow(dr)
                            End If
                        End If
                        subkey.Close()
                    Next

                    DefaultUserProfile = key.GetValue("DefaultUserProfile", "Default User").ToString
                    key.Close()

                    ' Update the PC table's NumUser field
                    If ds.PC.Count > 0 Then
                        ds.PC(0).PC_NumUsers = num_users
                    End If

                    ' Manually create an entry for the Default User.  This is
                    ' required by the Startup table as a foriegn key. 
                    If ds.Profiles.Count > 0 Then
                        Dim parts() As String

                        dr = ds.Profiles.NewProfilesRow
                        dr.ID_Profile = ID
                        dr.Profile = "Default User"

                        ' use an existing entry as a prototype
                        parts = ds.Profiles(0).Profile_Path.Split("\"c)
                        parts(parts.Length - 1) = DefaultUserProfile
                        ImagePath = Join(parts, "\")
                        dr.Profile_Path = Left(ImagePath, ds.Profiles.Profile_PathColumn.MaxLength)
                        dr.Profile_IsAdmin = False
                        ds.Profiles.AddProfilesRow(dr)

                        ' mount this hive?
                        If admin AndAlso Array.IndexOf(LoadedHives, DefaultUserProfile) < 0 Then
                            Try
                                Misc.MountHive(RemotePC, DefaultUserProfile, ImagePath & "\NTUSER.DAT")
                                ' we tag the MountPoint with an asterkisk
                                ' if we mounted it (used in the Dispose to
                                ' unmount the hives)
                                ProfileList.Add("Default User", "*" & DefaultUserProfile)
                            Catch ex As Exception
                                ' ignore
                            End Try
                        End If

                        ' Add this Profile/MountPoint to the list (it's not
                        ' likely that there would be a mounted Default User
                        ' hive unless we mounted it, but you never know...)
                        If Not ProfileList.ContainsKey("Default User") Then
                            ProfileList.Add("Default User", DefaultUserProfile)
                        End If
                    End If
                End If

                ' Let's do a "test run" to verify that the hive got mounted
                ' and that current user has permission to read it.  We do this
                ' here rather than in each one of the "dependent" methods.
                Dim PendingRemoval As New StringCollection

                For Each ProfileName In ProfileList.Keys
                    MountPoint = ProfileList(ProfileName).TrimStart("*"c)
                    Try
                        testkey = reg_hku.OpenSubKey(MountPoint & "\Control Panel", False)
                        If IsNothing(testkey) Then
                            ' every mounted hive should have a Control Panel
                            ' subkey.  If not, then something is wrong.
                            PendingRemoval.Add(ProfileName)
                        Else
                            testkey.Close()
                        End If
                    Catch
                        ' If this fails (typically due to a SecurityException),
                        ' then this key should be removed from the ProfileList
                        PendingRemoval.Add(ProfileName)
                    End Try
                Next

                ' Attempting to remove an item in a collection while inside
                ' a "for each" loop causes a InvalidOperationException.  So,
                ' we make a list of items to remove and then remove 'em.
                For Each ProfileName In PendingRemoval
                    ProfileList.Remove(ProfileName)
                Next
            Catch ex As Exception
                dr = ds.Profiles.NewProfilesRow
                dr.ID_Profile = ID
                dr.Profile = "Error"
                dr.Profile_IsAdmin = False
                dr.Profile_Path = Left("Error: " & ex.Message, ds.Profiles.Profile_PathColumn.MaxLength)
                ds.Profiles.AddProfilesRow(dr)
                Misc.ErrorLog(ex, MachineName, "Profiles", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' QFE (Quick Fix Engineering) = Windows updates
    ' The QFE feature of WMI is broken... so we have to do it hard way
    '
    Public Sub GetQFE()
        Dim dr As SOSOSDataset.QFERow

        Try
            ' We might find some of the entries in the "old location" and
            ' some of them in the "new location".  We rely on the "no
            ' duplicates" feature in each GetQFE routines to make sure
            ' we don't double up.  Since GetQFE_2k has the better info,
            ' we run it first
            Misc.GetQFE_2k(ds, reg_hklm, "SOFTWARE\Microsoft\Updates", RegWow64Options.KEY_WOW64_32KEY)
            If Is64bit Then
                Misc.GetQFE_2k(ds, reg_hklm, "SOFTWARE\Microsoft\Updates", RegWow64Options.KEY_WOW64_64KEY)
            End If
            Misc.GetQFE_9x(ds, reg_hklm, "SOFTWARE\Microsoft\Windows\CurrentVersion\Setup\Updates")
            Misc.GetQFE_NT(ds, reg_hklm, "SOFTWARE\Microsoft\Windows NT\CurrentVersion\HotFix")
            Misc.GetQFE_Office(ds, reg_hklm, "SOFTWARE\Microsoft\Windows\CurrentVersion\Installer\UserData\S-1-5-18\Products", RegWow64Options.KEY_WOW64_32KEY)
            If Is64bit Then
                Misc.GetQFE_Office(ds, reg_hklm, "SOFTWARE\Microsoft\Windows\CurrentVersion\Installer\UserData\S-1-5-18\Products", RegWow64Options.KEY_WOW64_64KEY)
            End If

            If OSVer >= 6.0 Then
                Misc.GetQFE_Vista(ds, Misc.ConvUNC(RemotePC, SystemRoot & "\Servicing\Packages"))
            End If

            ' Summary level info for the AutoUpdate table.
            ' QFE entries in older clients don't always support install
            ' dates, so let's not confuse things by showing the LastQFE
            ' based upon only those few entries that do have dates.
            If OSVer >= 5.0 Then
                Dim TempQFE As String
                Dim TempInstallDate As Date

                TempQFE = ""
                TempInstallDate = Date.MinValue
                For Each dr In ds.QFE
                    If Not dr.IsQFE_InstallDateNull Then
                        If TempInstallDate < dr.QFE_InstallDate Then
                            TempQFE = dr.QFE
                            TempInstallDate = dr.QFE_InstallDate
                        End If
                    End If
                Next
                If TempQFE <> "" And ds.AutoUpdate.Count > 0 Then
                    ' make sure we don't obliterate an error message
                    If ds.AutoUpdate(0).IsAutoUpdate_LastQFENull Then
                        ds.AutoUpdate(0).AutoUpdate_LastQFE = TempQFE
                        ds.AutoUpdate(0).AutoUpdate_LastQFEDate = TempInstallDate
                    End If
                End If
            End If

            ' Note: We do a bit more processing of the QFE Table in
            ' SoftwareSummary
        Catch ex As Exception
            dr = ds.QFE.NewQFERow
            dr.ID_QFE = ID
            dr.QFE = Left("Error: " & ex.Message, ds.QFE.QFEColumn.MaxLength)
            ds.QFE.AddQFERow(dr)
            Misc.ErrorLog(ex, MachineName, "QFE", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the Registry settings for the ad-hoc list of registry values
    ' Note: The list of keys comes from the *.config file
    ' Note: Duplicate entries in the RegistryList cause exceptions
    '
    Public Sub GetRegistry()
        Dim key_temp(), key_list, key_root, key_value, key_name, value As String
        Dim key, subkey As RegistryKey
        Dim dr As SOSOSDataset.RegistryRow

        Try
            For Each key_list In My.Settings.RegistryList
                If key_list <> "" Then
                    ' parse the key_list value into "root" "name", and "value"
                    key_temp = Split(key_list, "\")
                    key_root = key_temp(0)
                    key_value = key_temp(UBound(key_temp))
                    If key_list.Length - key_root.Length - key_value.Length - 2 > 0 Then
                        key_name = key_list.Substring(key_root.Length + 1, key_list.Length - key_value.Length - key_root.Length - 2)

                        ' open the correct root key
                        key = reg_hklm
                        Select Case key_root.ToUpper
                            Case "HKEY_LOCAL_MACHINE", "HKLM"
                                key = reg_hklm
                            Case "HKEY_CLASSES_ROOT", "HKCR"
                                key = reg_hkcr
                            Case "HKEY_CURRENT_USER", "HKCU"
                                If RemotePC <> "" Then
                                    key = RegistryKey.OpenRemoteBaseKey(RegistryHive.CurrentUser, RemotePC)
                                Else
                                    key = Registry.CurrentUser
                                End If
                            Case "HKEY_USERS", "HKU"
                                key = reg_hku
                            Case "HKEY_CURRENT_CONFIG", "HKCC"
                                If RemotePC <> "" Then
                                    key = RegistryKey.OpenRemoteBaseKey(RegistryHive.CurrentConfig, RemotePC)
                                Else
                                    key = Registry.CurrentConfig
                                End If
                            Case Else
                                dr = ds.Registry.NewRegistryRow()
                                dr.ID_Registry = ID
                                dr.Registry_Key = Left("Error: Badly formatted key '" & key_list & "'", ds.Registry.Registry_KeyColumn.MaxLength)
                                ds.Registry.AddRegistryRow(dr)
                        End Select

                        ' open the key, get the value
                        value = ""
                        subkey = key.OpenSubKey(key_name, False)
                        If Not IsNothing(subkey) Then
                            Dim value_object As Object

                            value_object = subkey.GetValue(key_value)
                            If Not IsNothing(value_object) Then
                                value = value_object.ToString
                            End If

                            ' convert a System.Btye[] into something more useful
                            If value = "System.Byte[]" Then
                                Dim b() As Byte
                                b = CType(subkey.GetValue(key_value), Byte())
                                value = Misc.ByteArrayToString(b)
                            End If
                            subkey.Close()
                        End If

                        ' let's try that again using the virtualized registry key
                        If value = "" And Is64bit Then
                            subkey = key.OpenSubKey(key_name, False, RegWow64Options.KEY_WOW64_32KEY)
                            If Not IsNothing(subkey) Then
                                Dim value_object As Object

                                value_object = subkey.GetValue(key_value)
                                If Not IsNothing(value_object) Then
                                    value = value_object.ToString
                                End If

                                ' convert a System.Btye[] into something more useful
                                If value = "System.Byte[]" Then
                                    Dim b() As Byte
                                    b = CType(subkey.GetValue(key_value), Byte())
                                    value = Misc.ByteArrayToString(b)
                                End If
                                subkey.Close()
                            End If
                        End If

                        If value <> "" Then
                            dr = ds.Registry.NewRegistryRow()
                            dr.ID_Registry = ID
                            dr.Registry_Key = Left(key_list, ds.Registry.Registry_KeyColumn.MaxLength)
                            dr.Registry_Value = Left(value, ds.Registry.Registry_ValueColumn.MaxLength)
                            ds.Registry.AddRegistryRow(dr)
                        End If
                    End If
                End If
            Next
        Catch ex As Exception
            dr = ds.Registry.NewRegistryRow()
            dr.ID_Registry = ID
            dr.Registry_Key = Left("Error: " & ex.Message, ds.Registry.Registry_KeyColumn.MaxLength)
            ds.Registry.AddRegistryRow(dr)
            Misc.ErrorLog(ex, MachineName, "Registry", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the Windows Services
    '
    Public Sub GetServices()
        ' Not available on Win9x
        If IsWin32 Then
            Dim obj As ManagementObject
            Dim dr As SOSOSDataset.ServicesRow

            Try
                wmi.Path.RelativePath = "Win32_Service"
                For Each obj In wmi.GetInstances()
                    dr = ds.Services.NewServicesRow()
                    dr.ID_Service = ID
                    dr.Service = Left(obj("Name").ToString, ds.Services.ServiceColumn.MaxLength)
                    If Not IsNothing(obj("PathName")) Then
                        dr.Service_Path = Left(obj("PathName").ToString, ds.Services.Service_PathColumn.MaxLength)
                    End If
                    dr.Service_Status = Left(obj("State").ToString, ds.Services.Service_StatusColumn.MaxLength)
                    dr.Service_StartType = Left(obj("StartMode").ToString, ds.Services.Service_StartTypeColumn.MaxLength)
                    If Not IsNothing(obj("StartName")) Then
                        dr.Service_AcctName = Left(obj("StartName").ToString, ds.Services.Service_AcctNameColumn.MaxLength)
                    End If
                    ds.Services.AddServicesRow(dr)
                Next
            Catch ex As Exception
                dr = ds.Services.NewServicesRow()
                dr.ID_Service = ID
                dr.Service = Left("Error: " & ex.Message, ds.Services.ServiceColumn.MaxLength)
                ds.Services.AddServicesRow(dr)
                Misc.ErrorLog(ex, MachineName, "Services", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the list of shared directories and printers
    '
    Public Sub GetShares()
        Dim obj As ManagementObject
        Dim dr As SOSOSDataset.SharesRow

        Try
            wmi.Path.RelativePath = "Win32_Share"
            For Each obj In wmi.GetInstances()
                ' report only "non-administrative" shares
                If Not obj("Name").ToString.EndsWith("$") Then
                    dr = ds.Shares.NewSharesRow()
                    dr.ID_Share = ID
                    dr.Share_Desc = Left(obj("Name").ToString, ds.Shares.Share_DescColumn.MaxLength)
                    If Not IsNothing(obj("Path")) Then
                        dr.Share_Path = Left(obj("Path").ToString, ds.Shares.Share_PathColumn.MaxLength)
                    End If
                    If Not IsNothing(obj("Type")) Then
                        ' mask the upper bits
                        dr.Share_Type = Choose((Convert.ToInt32(obj("Type")) And &HFF) + 1, "Disk", "Printer", "Device", "IPC").ToString
                    End If
                    ds.Shares.AddSharesRow(dr)
                End If
            Next
        Catch ex As Exception
            dr = ds.Shares.NewSharesRow()
            dr.ID_Share = ID
            dr.Share_Desc = Left("Error: " & ex.Message, ds.Shares.Share_DescColumn.MaxLength)
            ds.Shares.AddSharesRow(dr)
            Misc.ErrorLog(ex, MachineName, "Shares", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' The "Installed Software" feature of WMI is limited to only those
    ' installed via the Microsoft installer..  so, we have to poke around
    ' the registry
    '
    Public Sub GetSoftware()
        Dim key, subkey As RegistryKey
        Dim temp, dateraw, datebuf As String
        Dim dr As SOSOSDataset.SoftwareRow

        Try
            key = reg_hklm.OpenSubKey("Software\Microsoft\Windows\CurrentVersion\Uninstall", False, RegWow64Options.KEY_WOW64_32KEY)
            If Not IsNothing(key) Then
                For Each temp In key.GetSubKeyNames()
                    subkey = key.OpenSubKey(temp, False, RegWow64Options.KEY_WOW64_32KEY)
                    If Not IsNothing(subkey.GetValue("DisplayName")) Then
                        dr = ds.Software.NewSoftwareRow
                        dr.ID_Software = ID
                        dr.Software = Left(subkey.GetValue("DisplayName").ToString, ds.Software.SoftwareColumn.MaxLength)
                        If Not IsNothing(subkey.GetValue("VersionMajor")) And Not IsNothing(subkey.GetValue("VersionMinor")) Then
                            dr.Software_Version = Left(subkey.GetValue("VersionMajor").ToString & "." & subkey.GetValue("VersionMinor").ToString, ds.Software.Software_VersionColumn.MaxLength)
                        End If
                        ' we prefer this one (so it's done last)
                        If Not IsNothing(subkey.GetValue("DisplayVersion")) Then
                            dr.Software_Version = Left(subkey.GetValue("DisplayVersion").ToString, ds.Software.Software_VersionColumn.MaxLength)
                        End If
                        If Not IsNothing(subkey.GetValue("Publisher")) Then
                            dr.Software_Vendor = Left(subkey.GetValue("Publisher").ToString, ds.Software.Software_VendorColumn.MaxLength)
                        End If
                        If Not IsNothing(subkey.GetValue("InstallDate")) Then
                            dateraw = subkey.GetValue("InstallDate").ToString
                            If IsDate(dateraw) Then
                                dr.Software_InstallDate = [Date].Parse(dateraw, culture)
                            ElseIf dateraw.Length = 8 Then
                                ' The normal format in the registry is yyyyMMdd, so
                                ' we're using the en-US format of MM/dd/yyyy for the
                                ' date conversion
                                datebuf = dateraw.Substring(4, 2) & "/" & dateraw.Substring(6, 2) & "/" & dateraw.Substring(0, 4)
                                If IsDate(datebuf) Then
                                    dr.Software_InstallDate = [Date].Parse(datebuf, culture)
                                End If
                            End If
                        End If
                        If Not IsNothing(subkey.GetValue("InstallSource")) Then
                            dr.Software_InstallSrc = Left(subkey.GetValue("InstallSource").ToString, ds.Software.Software_InstallSrcColumn.MaxLength)
                        End If

                        ' just ignore duplicate records 
                        If IsNothing(ds.Software.FindByID_SoftwareSoftware(dr.ID_Software, dr.Software)) Then
                            ds.Software.AddSoftwareRow(dr)
                        End If
                        subkey.Close()
                    End If
                Next
                key.Close()
            End If

            ' do it again for the 64-bit world
            If Is64bit Then
                key = reg_hklm.OpenSubKey("Software\Microsoft\Windows\CurrentVersion\Uninstall", False, RegWow64Options.KEY_WOW64_64KEY)
                If Not IsNothing(key) Then
                    For Each temp In key.GetSubKeyNames()
                        subkey = key.OpenSubKey(temp, False, RegWow64Options.KEY_WOW64_64KEY)
                        If Not IsNothing(subkey.GetValue("DisplayName")) Then
                            dr = ds.Software.NewSoftwareRow
                            dr.ID_Software = ID
                            dr.Software = Left(subkey.GetValue("DisplayName").ToString, ds.Software.SoftwareColumn.MaxLength)
                            If Not IsNothing(subkey.GetValue("VersionMajor")) And Not IsNothing(subkey.GetValue("VersionMinor")) Then
                                dr.Software_Version = Left(subkey.GetValue("VersionMajor").ToString & "." & subkey.GetValue("VersionMinor").ToString, ds.Software.Software_VersionColumn.MaxLength)
                            End If
                            ' we prefer this one (so it's done last)
                            If Not IsNothing(subkey.GetValue("DisplayVersion")) Then
                                dr.Software_Version = Left(subkey.GetValue("DisplayVersion").ToString, ds.Software.Software_VersionColumn.MaxLength)
                            End If
                            If Not IsNothing(subkey.GetValue("Publisher")) Then
                                dr.Software_Vendor = Left(subkey.GetValue("Publisher").ToString, ds.Software.Software_VendorColumn.MaxLength)
                            End If
                            If Not IsNothing(subkey.GetValue("InstallDate")) Then
                                dateraw = subkey.GetValue("InstallDate").ToString
                                If IsDate(dateraw) Then
                                    dr.Software_InstallDate = [Date].Parse(dateraw, culture)
                                ElseIf dateraw.Length = 8 Then
                                    ' The normal format in the registry is yyyyMMdd, so
                                    ' we're using the en-US format of MM/dd/yyyy for the
                                    ' date conversion
                                    datebuf = dateraw.Substring(4, 2) & "/" & dateraw.Substring(6, 2) & "/" & dateraw.Substring(0, 4)
                                    If IsDate(datebuf) Then
                                        dr.Software_InstallDate = [Date].Parse(datebuf, culture)
                                    End If
                                End If
                            End If
                            If Not IsNothing(subkey.GetValue("InstallSource")) Then
                                dr.Software_InstallSrc = Left(subkey.GetValue("InstallSource").ToString, ds.Software.Software_InstallSrcColumn.MaxLength)
                            End If

                            ' just ignore duplicate records 
                            If IsNothing(ds.Software.FindByID_SoftwareSoftware(dr.ID_Software, dr.Software)) Then
                                ds.Software.AddSoftwareRow(dr)
                            End If
                            subkey.Close()
                        End If
                    Next
                    key.Close()
                End If
            End If

            ' don't forget to make it look pretty
            Misc.DataTableSort(ds.Tables("Software"), "Software")
        Catch ex As Exception
            dr = ds.Software.NewSoftwareRow
            dr.ID_Software = ID
            dr.Software = Left("Error: " & ex.Message, ds.Software.SoftwareColumn.MaxLength)
            ds.Software.AddSoftwareRow(dr)
            Misc.ErrorLog(ex, MachineName, "Software", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the version numbers for Microsoft Word and Internet Explorer.
    ' Also performs "mining" of QFE data in the existing tables
    ' Note: Data is saved into existing tables, there is no separate
    ' SoftwareSummary table
    '
    Public Sub GetSoftwareSummary()
        Dim key As RegistryKey
        Dim temp, Word_Path, kb As String
        Dim cnt As Integer
        Dim dr_qfe As SOSOSDataset.QFERow
        Dim dr_cmp As SOSOSDataset.ComponentsRow
        Dim dr_sfw As SOSOSDataset.SoftwareRow

        Try
            ' Microsoft Office Version (actually just Word... but that's close enough)
            temp = ""
            Word_Path = ""
            key = reg_hkcr.OpenSubKey("CLSID\{00020900-0000-0000-C000-000000000046}\DefaultIcon", False, RegWow64Options.KEY_WOW64_32KEY)
            If Not IsNothing(key) Then
                temp = key.GetValue("", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString.Split(","c)(0)
                temp = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
                Word_Path = Misc.ConvUNC(RemotePC, temp)
                If File.Exists(Word_Path) Then
                    ds.SOS(0).MSOffice_Version = Misc.GetVersion(Word_Path)
                End If
                key.Close()
            End If

            ' Microsoft Office versions are complex... it's a combination of the
            ' application version numer and the office suite  version number.
            ' We get the suite number from the MSO.DLL or WWLIB.DLL file
            If Not ds.SOS(0).IsMSOffice_VersionNull And Word_Path.Length > 2 Then
                Dim Word_Version(), MSO_Path, MSO_Temp, MSO_Version() As String

                Word_Version = ds.SOS(0).MSOffice_Version.Split("."c)
                If UBound(Word_Version) >= 2 Then

                    Select Case CInt(Word_Version(0))
                        Case Is >= 12
                            ' for Office 2007 (and above?)
                            MSO_Path = Word_Path.ToLower.Replace("winword.exe", "wwlib.dll")
                            MSO_Temp = Misc.GetVersion(MSO_Path)
                            If MSO_Temp <> "" Then
                                MSO_Version = MSO_Temp.Split("."c)
                                If UBound(MSO_Version) >= 3 Then
                                    ds.SOS(0).MSOffice_Version = Word_Version(0) & "." & Word_Version(1) & "." & MSO_Version(2) & "." & MSO_Version(3)
                                End If
                            End If
                        Case 10, 11
                            ' for Office XP and Office 2003
                            MSO_Path = ""
                            key = reg_hkcr.OpenSubKey("odcnewfile\DefaultIcon", False)
                            If Not IsNothing(key) Then
                                temp = key.GetValue("", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString.ToString.Split(","c)(0)
                                temp = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
                                MSO_Path = Misc.ConvUNC(RemotePC, temp)
                                key.Close()
                            End If

                            If MSO_Path <> "" AndAlso File.Exists(MSO_Path) Then
                                MSO_Temp = Misc.GetVersion(MSO_Path)
                                If MSO_Temp <> "" Then
                                    MSO_Version = MSO_Temp.Split("."c)
                                    If UBound(MSO_Version) >= 2 Then
                                        ds.SOS(0).MSOffice_Version = Word_Version(0) & "." & Word_Version(2) & "." & MSO_Version(2)
                                    End If
                                End If
                            End If
                        Case Is < 10
                            ' for Office 2k and below
                            MSO_Path = Word_Path.ToLower.Replace("winword.exe", "mso") & Word_Version(0) & ".dll"
                            MSO_Temp = Misc.GetVersion(MSO_Path)
                            If MSO_Temp <> "" Then
                                ds.SOS(0).MSOffice_Version = MSO_Temp
                            End If
                    End Select
                End If
            End If

            ' IE Version (more straight forward)
            key = reg_hkcr.OpenSubKey("CLSID\{25336920-03F9-11CF-8FD0-00AA00686F13}\DefaultIcon", False)
            If Not IsNothing(key) Then
                Dim IE_Path As String
                temp = key.GetValue("", "", RegistryValueOptions.DoNotExpandEnvironmentNames).ToString.Split(","c)(0)
                temp = Misc.ExpandEnv(RemotePC, temp, SystemPathVars)
                IE_Path = Misc.ConvUNC(RemotePC, temp)
                If File.Exists(IE_Path) Then
                    temp = Misc.GetVersion(IE_Path)
                    ' shorten it up a wee bit
                    cnt = InStr(temp, " ")
                    If cnt > 0 Then
                        ds.SOS(0).IE_Version = Left(temp, cnt - 1)
                    Else
                        ds.SOS(0).IE_Version = temp
                    End If
                End If
                key.Close()
            End If

            ' Sometimes the Quick Fix Engineering (QFE) values are scattered 
            ' througout the other tables, so we go snooping for entries that
            ' need to be added to the QFE table.

            ' look in the Components table for potential QFEs
            For Each dr_cmp In ds.Components
                temp = dr_cmp.Component_Name
                If temp Like "*KB######*" Or temp Like "*Q######*" Or temp Like "*S######*" Then
                    dr_qfe = ds.QFE.NewQFERow
                    dr_qfe.ID_QFE = ID
                    dr_qfe.QFE = Left(temp, ds.QFE.QFEColumn.MaxLength)
                    dr_qfe.QFE_Description = Left(dr_cmp.Component, ds.QFE.QFE_DescriptionColumn.MaxLength)
                    dr_qfe.QFE_Category = "HotFix"
                    ' ignore duplicates
                    If IsNothing(ds.QFE.FindByID_QFEQFE(dr_qfe.ID_QFE, dr_qfe.QFE)) Then
                        ds.QFE.AddQFERow(dr_qfe)
                    End If
                End If
            Next

            ' look in the Software table for potential QFEs
            For Each dr_sfw In ds.Software
                temp = dr_sfw.Software
                kb = ""
                If temp Like "*KB######*" Then
                    kb = temp.Substring(temp.IndexOf("KB"), 8)
                End If
                If temp Like "*Q######*" Then
                    kb = temp.Substring(temp.IndexOf("Q"), 7)
                End If
                If kb <> "" Then
                    dr_qfe = ds.QFE.NewQFERow
                    dr_qfe.ID_QFE = ID
                    dr_qfe.QFE = kb
                    dr_qfe.QFE_Description = Left(temp, ds.QFE.QFE_DescriptionColumn.MaxLength)
                    dr_qfe.QFE_Category = "HotFix"
                    If Not dr_sfw.IsSoftware_InstallDateNull Then
                        dr_qfe.QFE_InstallDate = dr_sfw.Software_InstallDate
                    End If
                    ' ignore duplicates
                    If IsNothing(ds.QFE.FindByID_QFEQFE(dr_qfe.ID_QFE, dr_qfe.QFE)) Then
                        ds.QFE.AddQFERow(dr_qfe)
                    End If
                End If
            Next

            ' Sort 'em (again)
            Misc.DataTableSort(ds.QFE, "QFE")
        Catch ex As Exception
            ds.SOS(0).MSOffice_Version = Left("Error:" & ex.Message, ds.SOS.MSOffice_VersionColumn.MaxLength)
            Misc.ErrorLog(ex, MachineName, "SoftwareSummary", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get information about the User's startup settings
    '
    Public Sub GetStartup()
        Dim ProfileName, Profile, StartDir, MountPoint, value, temp As String
        Dim key As RegistryKey
        Dim cnt As Integer
        Dim shell As IWshRuntimeLibrary.IWshShell_Class
        Dim shortcut As IWshRuntimeLibrary.IWshShortcut_Class
        Dim dr As SOSOSDataSet.StartupRow

        ProfileName = ""
        Try
            ' let's make sure there is a cooresponding entry in the
            ' profiles table... if not, we create one
            If IsNothing(ds.Profiles.FindByID_ProfileProfile(ID, "Default User")) Then
                Dim drp As SOSOSDataSet.ProfilesRow
                drp = ds.Profiles.NewProfilesRow
                drp.ID_Profile = ID
                drp.Profile = "Default User"
                drp.Profile_IsAdmin = False
                ds.Profiles.AddProfilesRow(drp)
            End If

            shell = New IWshRuntimeLibrary.IWshShell_Class
            cnt = 0

            ' Record the common entries as part of the Default User
            ' profile.  This is not technically correct, since the
            ' Default profile has nothing to do with these registry
            ' settings. But it a reasonable way to reduce the number
            ' of duplicate entries for each of the other users.

            ' look in registry for HKLM's Run section
            key = reg_hklm.OpenSubKey("SOFTWARE\Microsoft\Windows\CurrentVersion\Run", False, RegWow64Options.KEY_WOW64_32KEY)
            If Not IsNothing(key) Then
                For Each value In key.GetValueNames
                    If value <> "" Then
                        dr = ds.Startup.NewStartupRow
                        dr.ID_Startup = ID
                        dr.ID_Profile = "Default User"
                        dr.Startup_ID = cnt
                        dr.Startup_Location = "HKLM Run"
                        dr.Startup_Name = Left(value, ds.Startup.Startup_NameColumn.MaxLength)
                        temp = key.GetValue(value, "").ToString
                        If temp <> "" Then
                            dr.Startup_Path = Left(temp, ds.Startup.Startup_PathColumn.MaxLength)
                        End If
                        ds.Startup.AddStartupRow(dr)
                        cnt += 1
                    End If
                Next
                key.Close()
            End If

            If Directory.Exists(Misc.ConvUNC(RemotePC, AllUsersProfile & "\Start Menu\Programs\Startup")) Then
                For Each temp In Directory.GetFiles(Misc.ConvUNC(RemotePC, AllUsersProfile & "\Start Menu\Programs\Startup"))
                    If Not temp.EndsWith(".ini") Then
                        dr = ds.Startup.NewStartupRow
                        dr.ID_Startup = ID
                        dr.ID_Profile = "Default User"
                        dr.Startup_ID = cnt
                        dr.Startup_Location = "Common Startup"
                        dr.Startup_Name = Left(Path.GetFileNameWithoutExtension(temp), ds.Startup.Startup_NameColumn.MaxLength)
                        If Not temp.EndsWith(".lnk") Then
                            ' special processing for bat files, we also
                            ' get the file size and mod date
                            If temp.ToLower.EndsWith(".bat") Then
                                Dim fi As FileInfo
                                Dim buf As String

                                fi = New FileInfo(temp)
                                buf = Path.GetFileName(temp)
                                buf &= ", Size=" & fi.Length
                                buf &= ", ModDate=" & fi.LastWriteTime
                                dr.Startup_Path = Left(buf, ds.Startup.Startup_PathColumn.MaxLength)
                            Else
                                dr.Startup_Path = Left(Path.GetFileName(temp), ds.Startup.Startup_PathColumn.MaxLength)
                            End If
                        Else
                            ' look inside the shortcut to get the
                            ' application's path
                            shortcut = CType(shell.CreateShortcut(temp), IWshRuntimeLibrary.IWshShortcut_Class)
                            dr.Startup_Path = Left(shortcut.TargetPath, ds.Startup.Startup_PathColumn.MaxLength)
                            shortcut = Nothing
                        End If
                        ds.Startup.AddStartupRow(dr)
                        cnt += 1
                    End If
                Next
            End If

            ' Now process the items for the real profiles
            For Each Profile In ProfileList.Keys
                ProfileName = Left(Profile, ds.Startup.ID_ProfileColumn.MaxLength)
                MountPoint = ProfileList(Profile).TrimStart("*"c)
                cnt = 0

                ' skip the Default User (we're already using that spot for
                ' the All Users data above)
                If String.Compare(Profile, "Default User", True) <> 0 Then
                    ' look in registry for User's HKCU Run section
                    key = reg_hku.OpenSubKey(MountPoint & "\SOFTWARE\Microsoft\Windows\CurrentVersion\Run", False)
                    If Not IsNothing(key) Then
                        For Each value In key.GetValueNames
                            If value <> "" Then
                                dr = ds.Startup.NewStartupRow
                                dr.ID_Startup = ID
                                dr.ID_Profile = ProfileName
                                dr.Startup_ID = cnt
                                dr.Startup_Location = "HKCU Run"
                                dr.Startup_Name = Left(value, ds.Startup.Startup_NameColumn.MaxLength)
                                temp = key.GetValue(value, "").ToString
                                If temp <> "" Then
                                    dr.Startup_Path = Left(temp, ds.Startup.Startup_PathColumn.MaxLength)
                                End If
                                ds.Startup.AddStartupRow(dr)
                                cnt += 1
                            End If
                        Next
                        key.Close()
                    End If

                    ' Look in the User's start menu.  We start by getting
                    ' the path to the profile from the Profiles table
                    Dim dr_profile As SOSOSDataSet.ProfilesRow = ds.Profiles.FindByID_ProfileProfile(ID, Profile)
                    StartDir = dr_profile.Profile_Path & "\Start Menu\Programs\Startup"
                    If Directory.Exists(Misc.ConvUNC(RemotePC, StartDir)) Then
                        For Each temp In Directory.GetFiles(Misc.ConvUNC(RemotePC, StartDir))
                            If Not temp.EndsWith(".ini") Then
                                dr = ds.Startup.NewStartupRow
                                dr.ID_Startup = ID
                                dr.ID_Profile = ProfileName
                                dr.Startup_ID = cnt
                                dr.Startup_Location = "Users Startup"
                                dr.Startup_Name = Left(Path.GetFileNameWithoutExtension(temp), ds.Startup.Startup_NameColumn.MaxLength)
                                If Not temp.EndsWith(".lnk") Then
                                    ' special processing for bat files, we also
                                    ' get the file size and mod date
                                    If temp.ToLower.EndsWith(".bat") Then
                                        Dim fi As FileInfo
                                        Dim buf As String

                                        fi = New FileInfo(temp)
                                        buf = Path.GetFileName(temp)
                                        buf &= ", Size=" & fi.Length
                                        buf &= ", ModDate=" & fi.LastWriteTime
                                        dr.Startup_Path = Left(buf, ds.Startup.Startup_PathColumn.MaxLength)
                                    Else
                                        dr.Startup_Path = Left(Path.GetFileName(temp), ds.Startup.Startup_PathColumn.MaxLength)
                                    End If
                                Else
                                    ' look inside the shortcut to get the
                                    ' application's path
                                    shortcut = CType(shell.CreateShortcut(temp), IWshRuntimeLibrary.IWshShortcut_Class)
                                    dr.Startup_Path = Left(shortcut.TargetPath, ds.Startup.Startup_PathColumn.MaxLength)
                                    shortcut = Nothing
                                End If
                                ds.Startup.AddStartupRow(dr)
                                cnt += 1
                            End If
                        Next
                    End If
                End If
            Next
            shell = Nothing
        Catch ex As Exception
            dr = ds.Startup.NewStartupRow
            dr.ID_Startup = ID
            ' it's possible that an early error wouldn't populate the
            ' ProfileName field (which is a part of the foreign key)
            If ProfileName = "" Then
                ProfileName = ds.Profiles(0).Profile
            End If
            dr.ID_Profile = ProfileName
            dr.Startup_ID = -1
            dr.Startup_Location = Left("Error: " & ex.Message, ds.Startup.Startup_LocationColumn.MaxLength)
            ds.Startup.AddStartupRow(dr)
            Misc.ErrorLog(ex, MachineName, "Startup", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the System Drivers (useful in the Recovery Console)
    '
    Public Sub GetSystemDrivers()
        ' Not available on Win9x
        If IsWin32 Then
            Dim obj As ManagementObject
            Dim dr As SOSOSDataset.SystemDriversRow

            Try
                wmi.Path.RelativePath = "Win32_SystemDriver"
                For Each obj In wmi.GetInstances()
#If TRIM_DISABLED_SYSDRIVERS Then
                    If obj("StartMode").ToString <> "Disabled" Then
#Else
                    If True then
#End If
                        dr = ds.SystemDrivers.NewSystemDriversRow
                        dr.ID_Driver = ID
                        dr.Driver_Name = Left(obj("Name").ToString, ds.SystemDrivers.Driver_NameColumn.MaxLength)
                        dr.Driver_Description = Left(obj("Description").ToString, ds.SystemDrivers.Driver_DescriptionColumn.MaxLength)
                        If Not IsNothing(obj("PathName")) Then
                            dr.Driver_Path = Left(obj("PathName").ToString, ds.SystemDrivers.Driver_PathColumn.MaxLength)
                        End If
                        dr.Driver_Type = Left(obj("ServiceType").ToString, ds.SystemDrivers.Driver_TypeColumn.MaxLength)
                        dr.Driver_StartMode = Left(obj("StartMode").ToString, ds.SystemDrivers.Driver_StartModeColumn.MaxLength)
                        ds.SystemDrivers.AddSystemDriversRow(dr)
                    End If
                Next
            Catch ex As Exception
                dr = ds.SystemDrivers.NewSystemDriversRow
                dr.ID_Driver = ID
                dr.Driver_Name = Left("Error: " & ex.Message, ds.SystemDrivers.Driver_NameColumn.MaxLength)
                ds.SystemDrivers.AddSystemDriversRow(dr)
                Misc.ErrorLog(ex, MachineName, "SystemDrivers", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get the status of the scheduled tasks
    ' Note: Only administrators can get task information
    '
    Public Sub GetTasks()
        Dim dr As SOSOSDataset.TasksRow

        Try
            Dim st As New ScheduledTasks

            If Environment.OSVersion.Version.Major >= 6 And OSVer >= 6.0 Then
                ' Sorry, you can only get Vista Scheduled Task when
                ' running on a Vista PC
                st.GetTaskInfoVista(MachineName, ID, ds.Tasks)
            Else
                ' The "mining" of scheduled tasks is complex enough to be 
                ' put in a separate class (lot's of API calls, etc)
                st.GetTaskInfo(MachineName, ID, SystemRoot, ds.Tasks)
            End If
        Catch ex As Exception
            dr = ds.Tasks.NewTasksRow
            dr.ID_Tasks = ID
            dr.Task_ID = -1
            dr.Task_JobName = Left(ex.Message, ds.Tasks.Task_JobNameColumn.MaxLength)
            ds.Tasks.AddTasksRow(dr)

            ' Record this only if it's an "unexpected" error
            If Not ex.GetType Is GetType(System.UnauthorizedAccessException) Then
                Misc.ErrorLog(ex, MachineName, "Tasks", Misc.ErrLogLevel.Errors)
            End If
        End Try
    End Sub

    '
    ' Record the User Login information (based upon data already collected)
    '
    Public Sub GetUserLogin()
        If UID <> "" Then
            Dim dr As SOSOSDataset.UserLoginRow
            Try
                dr = ds.UserLogin.NewUserLoginRow
                dr.UserID = UID
                dr.LoginTime = Now()
                dr.AcctDomain = Split(UID, "\")(0)
                dr.AcctName = Split(UID, "\")(1)
                If ds.PC.Count > 0 AndAlso Not ds.PC(0).IsPC_DomainNameNBNull Then
                    dr.ResourceDomain = ds.PC(0).PC_DomainNameNB
                End If
                dr.PCName = MachineName
                ' not the primary key this time
                dr.ID = ID
                ds.UserLogin.AddUserLoginRow(dr)
            Catch ex As Exception
                ' It's possible to get database concurrency errors if the same 
                ' user logged on to multiple PC at exactly the same time.
                dr = ds.UserLogin.NewUserLoginRow
                dr.UserID = UID
                dr.LoginTime = Now()
                dr.AcctDomain = Left("Error: " & ex.Message, ds.UserLogin.AcctDomainColumn.MaxLength)
                ds.UserLogin.AddUserLoginRow(dr)
                Misc.ErrorLog(ex, MachineName, "UserLogin", Misc.ErrLogLevel.Errors)
            End Try
        End If
    End Sub

    '
    ' Get information about the Video Card
    ' Note: Non-Administrator's get minimal information
    '
    Public Sub GetVideo()
        Dim obj As ManagementObject
        Dim cnt As Integer
        Dim d As Double
        Dim dr As SOSOSDataset.VideoRow
        Dim PreviousVideo As String

        Try
            cnt = 0
            PreviousVideo = ""
            wmi.Path.RelativePath = "Win32_VideoController"
            For Each obj In wmi.GetInstances()
                dr = ds.Video.NewVideoRow
                dr.ID_Video = ID
                dr.Video_ID = cnt
                dr.Video = Left(obj("Caption").ToString, ds.Video.VideoColumn.MaxLength)
                If Not IsNothing(obj("CurrentRefreshRate")) Then
                    Try
                        d = 0
                        d = Convert.ToDouble(obj("CurrentRefreshRate"))
                    Catch
                    End Try
                    ' a value of 0 or 0xffffff means default or maximum
                    If d > 0 And d < 500 Then
                        dr.Video_Frequency = CInt(d)
                    End If
                End If
                If Not IsNothing(obj("CurrentScanMode")) Then
                    dr.Video_ScanMode = Choose(Convert.ToInt32(obj("CurrentScanMode")), "Other", "Unknown", "Interlaced", "Non-Interlaced").ToString
                End If
                If Not IsNothing(obj("DriverVersion")) Then
                    dr.Video_Driver = Left(obj("DriverVersion").ToString, ds.Video.Video_DriverColumn.MaxLength)
                End If
                If Not IsNothing(obj("AdapterRAM")) Then
                    dr.Video_AdapterRAM = CInt(Convert.ToDouble(obj("AdapterRAM")) / (1024.0# * 1024.0#))
                End If
                If Not IsNothing(obj("CurrentHorizontalResolution")) Then
                    dr.Video_Resolution = obj("CurrentHorizontalResolution").ToString & "x" & obj("CurrentVerticalResolution").ToString
                End If
                If Not IsNothing(obj("CurrentBitsPerPixel")) Then
                    dr.Video_ColorDepth = Convert.ToInt32(obj("CurrentBitsPerPixel"))
                End If

#If TRIM_PHANTOM_VIDEOS Then
                ' SMS "takes over" the WMI entry for the physical video card,
                ' so to overcome this, we just overwrite the SMS record with
                ' parts of the data from the physical video card
                If cnt > 0 And (PreviousVideo.EndsWith("Mirror Driver") Or PreviousVideo.EndsWith("Remote Control Driver")) Then
                    ds.Video(cnt - 1).Video = dr.Video
                    If dr.IsVideo_DriverNull Then
                        ds.Video(cnt - 1).SetVideo_DriverNull()
                    Else
                        ds.Video(cnt - 1).Video_Driver = dr.Video_Driver
                    End If
                    ds.Video(cnt - 1).Video_AdapterRAM = dr.Video_AdapterRAM
                End If

                ' Don't record "phantom" video cards (ones without a
                ' Video_Freqency)
                If cnt = 0 Or dr.IsVideo_FrequencyNull = False Then
                    ds.Video.AddVideoRow(dr)
                    PreviousVideo = dr.Video
                    cnt += 1
                End If
#Else
                ds.Video.AddVideoRow(dr)
                cnt += 1
#End If
            Next

            ' Check to see if we got anything... if not, then we try this 
            ' older approach.
            If ds.Video.Count = 0 Then
                cnt = 0
                PreviousVideo = ""
                wmi.Path.RelativePath = "Win32_VideoConfiguration"
                For Each obj In wmi.GetInstances()
                    dr = ds.Video.NewVideoRow
                    dr.ID_Video = ID
                    dr.Video_ID = cnt
                    dr.Video = Left(obj("AdapterDescription").ToString, ds.Video.VideoColumn.MaxLength)
                    If Not IsNothing(obj("RefreshRate")) Then
                        Try
                            d = 0
                            d = Convert.ToDouble(obj("RefreshRate"))
                        Catch
                        End Try
                        ' a value of 0 or 0xffffff means default or maximum
                        If d > 0 And d < 500 Then
                            dr.Video_Frequency = CInt(d)
                        End If
                    End If
                    If Not IsNothing(obj("ScanMode")) Then
                        dr.Video_ScanMode = Left(obj("ScanMode").ToString, ds.Video.Video_ScanModeColumn.MaxLength)
                    End If
                    dr.Video_Driver = Left(obj("InstalledDisplayDrivers").ToString, ds.Video.Video_DriverColumn.MaxLength)
                    If Not IsNothing(obj("AdapterRAM")) Then
                        dr.Video_AdapterRAM = CInt(Convert.ToDouble(obj("AdapterRAM")) / (1024.0# * 1024.0#))
                    End If
                    If Not IsNothing(obj("HorizontalResolution")) Then
                        dr.Video_Resolution = obj("HorizontalResolution").ToString & "x" & obj("VerticalResolution").ToString
                    End If
                    If Not IsNothing(obj("BitsPerPixel")) Then
                        dr.Video_ColorDepth = Convert.ToInt32(obj("BitsPerPixel"))
                    End If

#If TRIM_PHANTOM_VIDEOS Then
                    ' SMS "takes over" the WMI entry for the physical video card,
                    ' so to overcome this, we just overwrite the SMS record with
                    ' parts of the data from the physical video card
                    If cnt > 0 And (PreviousVideo.EndsWith("Mirror Driver") Or PreviousVideo.EndsWith("Remote Control Driver")) Then
                        ds.Video(cnt - 1).Video = dr.Video
                        If dr.IsVideo_DriverNull Then
                            ds.Video(cnt - 1).SetVideo_DriverNull()
                        Else
                            ds.Video(cnt - 1).Video_Driver = dr.Video_Driver
                        End If
                        ds.Video(cnt - 1).Video_AdapterRAM = dr.Video_AdapterRAM
                    End If

                    ' Don't record "phantom" video cards (ones without a
                    ' Video_Freqency)
                    If cnt = 0 Or dr.IsVideo_FrequencyNull = False Then
                        ds.Video.AddVideoRow(dr)
                        PreviousVideo = dr.Video
                        cnt += 1
                    End If
#Else
                    ds.Video.AddVideoRow(dr)
                    cnt += 1
#End If
                Next
            End If
        Catch ex As Exception
            dr = ds.Video.NewVideoRow
            dr.ID_Video = ID
            dr.Video_ID = -1
            dr.Video = Left("Error: " & ex.Message, ds.Video.VideoColumn.MaxLength)
            ds.Video.AddVideoRow(dr)
            Misc.ErrorLog(ex, MachineName, "Video", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Get the information from the Norton AntiVirus scan logs
    ' Note:  We use Norton AntiVirus Corporate Edition (NAVCE) so these
    ' routines are specific to that version.
    '
    Public Sub GetVirus()
        Dim log_dir, log_dirs(), log_file, temp() As String
        Dim year, mon, day, hr, min, sec, cnt As Integer
        Dim date_str, buf As String
        Dim found_it As Boolean
        Dim fs As FileStream
        Dim sr As StreamReader
        Dim dr As SOSOSDataset.VirusRow

        ' first thing we do... is find the log directory
        log_dirs = New String() {AllUsersProfile & "\Symantec\Symantec Endpoint Protection\Logs", _
                                AllUsersProfile & "\Application Data\Symantec\Symantec Endpoint Protection\Logs", _
                                AllUsersProfile & "\Application Data\Symantec\Norton AntiVirus Corporate Edition\7.5\Logs", _
                                AllUsersProfile & "\Application Data\Symantec\Symantec AntiVirus Corporate Edition\7.5\Logs", _
                                ProgramFiles & "\Symantec AntiVirus\logs", _
                                ProgramFiles & "\Norton AntiVirus\logs", _
                                ProgramFiles & "\NAVNT\Logs", _
                                ProgramFiles & "\NAV\logs"}

        found_it = False
        log_dir = ""
        For Each log_dir In log_dirs
            Try
                If Directory.Exists(Misc.ConvUNC(RemotePC, log_dir)) Then
                    found_it = True
                    Exit For
                End If
            Catch
            End Try
        Next
        ' sanity check
        If Not found_it Then
            dr = ds.Virus.NewVirusRow
            dr.ID_Virus = ID
            dr.Virus_ID = -1
            dr.Virus_Name = "Error: Can't locate log directory"
            ds.Virus.AddVirusRow(dr)

            ' Since this is common error on systems without NAVCE, we don't
            ' record it with the ErrorLog function
            Exit Sub
        End If

        ' Now we loop thru each log file in that directory
        cnt = 0
        For Each log_file In Directory.GetFiles(Misc.ConvUNC(RemotePC, log_dir), "*.log")
            Try
                ' open the log file
                fs = New FileStream(log_file, FileMode.Open, FileAccess.Read)
                sr = New StreamReader(fs)
                Do While sr.Peek() > -1
                    temp = Split(sr.ReadLine, ",")
#If TRIM_TAMPER_ATTEMPTS Then
                    If UBound(temp) >= 10 AndAlso (temp(1) = "5" Or (temp(1) = "14" And temp(13).IndexOf("err=") > 0)) Then
#Else
                    If UBound(temp) >= 10 AndAlso (temp(1) = "5" Or temp(1) = "45" Or (temp(1) = "14" And temp(13).IndexOf("err=") > 0)) Then
#End If
                        ' decode the timestamp
                        year = CInt("&H" & temp(0).Substring(0, 2))
                        mon = CInt("&H" & temp(0).Substring(2, 2))
                        day = CInt("&H" & temp(0).Substring(4, 2))
                        hr = CInt("&H" & temp(0).Substring(6, 2))
                        min = CInt("&H" & temp(0).Substring(8, 2))
                        sec = CInt("&H" & temp(0).Substring(10, 2))
                        date_str = year + 1970 & "/" & mon + 1 & "/" & day & " " & hr & ":" & min & ":" & sec

                        ' copy the fields that are of interest
                        dr = ds.Virus.NewVirusRow
                        dr.ID_Virus = ID
                        dr.Virus_ID = cnt
                        dr.Virus_Date = Date.Parse(date_str, culture)
                        dr.Virus_User = Left(temp(5), ds.Virus.Virus_UserColumn.MaxLength)

                        Select Case temp(1)
                            Case "14"
                                ' is a system type error (not really a virus)
                                dr.Virus_Name = Left(temp(13), ds.Virus.Virus_NameColumn.MaxLength)
                            Case "45"
                                ' is a tamper attempt (new for v10)
                                dr.Virus_Name = Left(temp(13), ds.Virus.Virus_NameColumn.MaxLength)
                                dr.Virus_Action = "Tamper Attempt"
                            Case Else
                                ' is a real virus
                                dr.Virus_Name = Left(temp(6), ds.Virus.Virus_NameColumn.MaxLength)
                                dr.Virus_Location = Left(temp(7), ds.Virus.Virus_LocationColumn.MaxLength)
                                ' take the first digit of a two-digit action code
                                buf = CType(Choose(CInt(temp(10).Substring(0, 1)), "Quarantined", "Renamed", "Deleted", "Left Alone", "Cleaned", "Cleaned or Macros Deleted"), String)
                                If IsNothing(buf) Then
                                    buf = "Unknown"
                                End If
                                dr.Virus_Action = buf
                        End Select
                        ds.Virus.AddVirusRow(dr)
                        cnt += 1
                    End If
                Loop
                sr.Close()
                fs.Close()
            Catch ex As Exception
                dr = ds.Virus.NewVirusRow
                dr.ID_Virus = ID
                dr.Virus_ID = -1
                dr.Virus_Name = Left("Error: " & ex.Message, ds.Virus.Virus_NameColumn.MaxLength)
                ds.Virus.AddVirusRow(dr)
                Misc.ErrorLog(ex, MachineName, "Virus", Misc.ErrLogLevel.Errors)
            End Try
        Next
    End Sub

    '
    ' Get the Norton AntiVirus "summary" information for the SOS table
    ' Note: records the data into the SOS table, there is no separate
    ' VirusSummary table.
    '
    Public Sub GetVirusSummary()
        Dim NortonDir(), NortonExe(), NortonDefs(), tempdir, temp, tdate, pname, defdir, d As String
        Dim fvi As FileVersionInfo
        Dim i, j As Integer
        Dim key As RegistryKey
        Dim buf(8) As Byte

        Try
            ' Get the Norton AnitVirus program version number
            NortonDir = New String() {ProgramFiles & "\Norton Antivirus\", _
                                     ProgramFiles & "\NAVNT\", _
                                     ProgramFiles & "\Norton SystemWorks\Norton Antivirus\", _
                                     ProgramFiles & "\Symantec_Client_Security\Symantec Antivirus\", _
                                     ProgramFiles & "\Symantec Client Security\Symantec Antivirus\", _
                                     ProgramFiles & "\Symantec Antivirus\", _
                                     ProgramFiles & "\Symantec\Symantec Endpoint Protection\", _
                                     ProgramFilesx86 & "\Symantec Antivirus\", _
                                     ProgramFilesx86 & "\Symantec\Symantec Endpoint Protection\"}

            NortonExe = New String() {"navw32.exe", "navwnt.exe", "vpc32.exe", "SymCorpUI.exe"}

            NortonDefs = New String() {AllUsersProfile & "\Symantec\Definitions\VirusDefs\", _
                                      CommonProgramFiles & "\Symantec Shared\VirusDefs\", _
                                      CommonProgramFilesx86 & "\Symantec Shared\VirusDefs\"}

            For i = 0 To UBound(NortonDir)
                For j = 0 To UBound(NortonExe)
                    temp = Misc.ConvUNC(RemotePC, NortonDir(i) & NortonExe(j))
                    If File.Exists(temp) Then
                        fvi = FileVersionInfo.GetVersionInfo(temp)
                        ds.SOS(0).Virus_Version = Left(fvi.FileVersion, ds.SOS.Virus_VersionColumn.MaxLength)
                        Exit For
                    End If
                Next
                ' have we found it?
                If Not ds.SOS(0).IsVirus_VersionNull Then
                    Exit For
                End If
            Next

            ' Get the date of the virus definitions
            defdir = ""
            For Each tempdir In NortonDefs
                pname = Misc.ConvUNC(RemotePC, tempdir)
                If Directory.Exists(pname) Then
                    For Each d In Directory.GetDirectories(pname)
                        ' find the last numeric directory
                        If IsNumeric(d.Substring(pname.Length)) Then
                            If d > defdir Then
                                defdir = d
                            End If
                        End If
                    Next
                    ' do the virus definition times like Norton does
                    temp = Path.GetFileName(defdir)
                    If temp.Length >= 8 Then
                        tdate = temp.Substring(4, 2) & "/" & temp.Substring(6, 2) & "/" & temp.Substring(0, 4)
                        If IsDate(tdate) Then
                            ds.SOS(0).Virus_Defs = Date.Parse(tdate, culture)
                        End If
                    End If
                    Exit For
                End If
            Next

            ' Get the date of the last virus scan 
            key = reg_hklm.OpenSubKey("SOFTWARE\Intel\LANDesk\VirusProtect6\CurrentVersion", False, RegWow64Options.KEY_WOW64_32KEY)
            If Not IsNothing(key) Then
                If Not IsNothing(key.GetValue("TimeOfLastScan")) Then
                    buf = CType(key.GetValue("TimeOfLastScan"), Byte())
                    temp = buf(1) + 1 & "/" & buf(2) & "/" & buf(0) + 1970 & " " & Format(buf(3), "00") & ":" & Format(buf(4), "00") & ":" & Format(buf(5), "00")
                    If IsDate(temp) Then
                        ds.SOS(0).Virus_Scan = Date.Parse(temp, culture)
                    End If
                End If
                key.Close()
            End If

            ' Newer Symantec Endpoint Protection (aka v11)
            If ds.SOS(0).IsVirus_ScanNull Then
                key = reg_hklm.OpenSubKey("SOFTWARE\Symantec\Symantec Endpoint Protection\AV", False, RegWow64Options.KEY_WOW64_32KEY)
                If Not IsNothing(key) Then
                    If Not IsNothing(key.GetValue("TimeOfLastScan")) Then
                        buf = CType(key.GetValue("TimeOfLastScan"), Byte())
                        temp = buf(1) + 1 & "/" & buf(2) & "/" & buf(0) + 1970 & " " & Format(buf(3), "00") & ":" & Format(buf(4), "00") & ":" & Format(buf(5), "00")
                        If IsDate(temp) Then
                            ds.SOS(0).Virus_Scan = Date.Parse(temp, culture)
                        End If
                    End If
                    key.Close()
                End If
            End If
        Catch ex As Exception
            ds.SOS(0).Virus_Version = Left("Error:" & ex.Message, ds.SOS.Virus_VersionColumn.MaxLength)
            Misc.ErrorLog(ex, MachineName, "VirusSummary", Misc.ErrLogLevel.Errors)
        End Try
    End Sub

    '
    ' Clean up before we go...  If we mounted a set of registry hives, then we
    ' should umount them.
    '
    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If Not Me.disposed Then
            If disposing Then
                Dim key, MountPoint As String
                Dim Recorded As Boolean

                If Not IsNothing(wmi) Then
                    wmi.Dispose()
                End If

                Recorded = False
                For Each key In ProfileList.Keys
                    MountPoint = ProfileList(key)
                    ' did we mount this registry hive?
                    If MountPoint.StartsWith("*") Then
                        Try
                            Misc.UnMountHive(RemotePC, MountPoint.TrimStart("*"c))
                        Catch ex87 As System.ComponentModel.Win32Exception When ex87.NativeErrorCode = 87
                            ' An error of "The parameter is incorrect" just
                            ' means that the system has already unloaded
                            ' the registry hive, so we can safely ignore
                            ' this error
                        Catch ex As Exception
                            ' Silently record this to the ErrorLog
                            If Not Recorded Then
                                Misc.ErrorLog(ex, MachineName, "UnmountHives", Misc.ErrLogLevel.Errors)
                            End If
                            Recorded = True
                        End Try
                    End If
                Next
            End If
        End If
        Me.disposed = True
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

    Protected Overrides Sub Finalize()
        Dispose(False)
        MyBase.Finalize()
    End Sub
End Class